/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: back_end.h
 *
 *  CONTENTS:  Main include file for BACK END , PSGRND.
 *             Classes: 'Grounder' - converts predicate theory to propositional
 *                      'GroundRules' - Performs propagation, lookahead, and
 *                                      output of the ground rules.
 *
 *  DATE:     May 8, 2004
 *
 *  MODIFIED BY: Artur Mikitiuk, Jul 22, 2004
 ***************************************************************************/

#include "binary.h"

// [1] Values that control which atom numbers will be assigned to particular
//  ground atoms and will be output in the ground theory for the solver:
	// The number to be assigned to the atoms that are true or false:
#define GR_ANUM_TRUEFALSE       0
	// The number to be assigned to the atoms that do
	//  not occur in the ground theory
#define GR_ANUM_NODIFF          0xFFFFFFFF
	// The upper limit for number assignment. Once the atom number
	//  is equal to 'GR_ANUM_ULIMIT', the grounder reports the
	//  error that there are not enough numbers to generate output
#define GR_ANUM_ULIMIT          0xFFFFFFFF


// [2] MEMORY MANAGEMENT CONSTANTS:

	// Default page size (number of elements)
	// This CANNOT be LESS THAN 10:
#define COMMON_PAGE_SIZE                100000
	// Compaction occurrs when the size of the allocated memory increses by
	//  this percent (should be not less than 1.1)
#define GR_COMPACTION_MULTIPLIER        1.5
	// Size of the ground theory (number of elements) when the first
	//  compaction occurrs:
	// Should be not less than 10 page sizes.
#define GR_MIN_COMPACTION_SIZE          1050000
	// When page needs to be expanded, this is ratio of the new page
	//  size to the old page size. Also, if some rule is copied from
	//  one page to another, the size of another page should be
	//  'PAGE_GROWTH_MUL' times as big as the rule part being copied.
	// NOTE: The next value should be integer!
#define PAGE_GROWTH_MUL                 2


//[3] 'GroundRules::controlFlag' can be any combination of the following values:
// (a) Each of these ENABLE some stage of ground theory processing:
#define GR_DO_PROPAGATE         1
#define GR_DO_LOOKAHEAD         2
	// Enable compaction of the whole theory after input of some rules
	//  when the size of the theory exceeds certain limits:
#define GR_DO_M_COMPACTION      4
	// Enable ccompaction of one rule after input of each rule. Enable
	//  compaction of the whole theory after the propagation and lookahead:
#define GR_DO_ANY_COMPACTION    8
	// Allows to see the sequence of atoms
	//  determined during propagation and lookahead:
#define GR_DO_SCRIPT            16
// (b) Other flag
	// Auxiliary flag for internal use by the lookahead:
#define GR_SCRIPT_LA            32
	// This flag indicates that the theory was found to be inconsistent:
#define GR_FLG_INCONSISTENT     64
	// Error in the ground rules: repetition in a
	//  weight atom with different weights
#define GR_FLG_ERROR            128

// Macros to work with 'GR_FLG_INCONSISTENT':
//   Check whether the inconsistency flag is set:
#define THEORY_IS_CONSISTENT (!(controlFlag & GR_FLG_INCONSISTENT))
//   Set the inconsistency flag:
#define INDICATE_INCONSISTENCY (controlFlag|=GR_FLG_INCONSISTENT)


// [4] CONSTANTS TO BE GIVEN IN THE ARGUMENTS TO METHODS:

//  (a) Constants that for parameter 'executionFlag' in the method
//  'GroundRules::executeFinalStage':
// See parameters for this method in the file "gr_rules.cpp" for details:
#define GROUT_ERRORS            1
#define GROUT_LEGACY            2
#define GROUT_DIMACS            4
#define GROUT_PBS               8

//  (b) Flags for the method 'GroundRules::compactRules'
//  See comment to the method for more details:
#define GR_CM_ALL               1
#define GR_CM_LASTRULE          2
#define GR_CM_ALLFINAL          3


// [5] DELIMITERS AND MACROS TO WORK WITH THEM:
//  See documentation for more information on the delimiters:
//    End-of-page delimiter
#define DELIM_PAGE              0
//    Beginning of a groud rule: implication
#define DELIM_IMPLY             1
//    Beginning of a groud rule: horn rule
#define DELIM_HORN              2
//    Separator between the body and the head (body is always the first):
#define DELIM_HEAD              3
//    Beginning of the weight atom:
#define DELIM_WEIGHT            4
//    Beginning of the cardinality atom:
#define DELIM_CARD              5
//    Weight value:
#define DELIM_WVAL              6
//  Total number of different delimiters:
#define DELIM_NUM               7

// Check whether the pointer is a delimiter, i.e. points to some
//  element in the array of delimiters:
#define IS_DELIM(ptr) ((ptr)>=delimiters && (ptr) < delimiters+DELIM_NUM)
// Get the number correspoding to a pointer which is delimiter
//  (corresponds to the constants DELIM_...):
#define GET_DELIM_TYPE(ptr) ((ptr)-delimiters)
//  Given the number of delimiter, this macro returns the
//   pointer corresponding to the delimiter:
#define MAKE_DELIM(num) (delimiters + (num))

// Checks whether the object (atom or delimiter) can occur in the atom list:
#define IN_ATOM_LIST(ptr) \
	((ptr) > delimiters+DELIM_CARD || (ptr) < delimiters+DELIM_PAGE)
// Checks whether the object (atom or delimiter) can occur INSIDE a rule:
#define IN_RULE(ptr) \
	((ptr) > delimiters + DELIM_HORN || (ptr) < delimiters + DELIM_PAGE)
// Opposite to 'IS_DELIM':
#define IS_ATOM(ptr) \
	((ptr) > delimiters + DELIM_WVAL || (ptr) < delimiters + DELIM_PAGE)

// Check whether the current delimiter delimits cardinality or weight atom
//  If so, we should be careful with the next element because it contains
//   the meta-limits that can be confused with delimiter or atom:
#define NEED_TO_SKIP_ONE(ptr) \
	((ptr) == delimiters+DELIM_CARD || (ptr) == delimiters+DELIM_WEIGHT)


// [6] VALUES OF GROUND ATOMS, META-ATOMS, AND RULES:
//    Default value:
#define GR_NORMAL               0
//   Lookahead flag. This may be combined with any other values and indicate
//     that the value is temporary, i.e. it was assigned during lookahead
//     and will be cleared shortly after:
#define GR_TEMP                 1
//    Program or meta atom is True:
#define GR_TRUE                 2
//    Program or meta atom is False:
#define GR_FALSE                4
//   Can be used only with horn atoms and indicates horn atoms
//     that were forced to be true by an implication:
#define GR_PSEUDOTRUE           8
//   The following value (in comment) can appear during lookahead when
//     the atom was assigned PSEUDO-TRUE during propagation and TRUE
//     during lookahead. Need to be very careful with this case:
//       GR_TRUE|GR_PSEUDOTRUE|GR_TEMP
//   The following value applies only to rules and
//     indicate that a rule is eliminated:
#define GR_ELIMINATED           16


	// The macro tells whether the atom should
	//  be considered to be True in an implication:
#define GR_IS_TRUE_I(value) ((value)&(GR_TRUE|GR_PSEUDOTRUE))
	// Self-explanatory:
#define GR_IS_TRUE(value) ((value)&GR_TRUE)
#define GR_IS_FALSE(value) ((value)&GR_FALSE)
#define GR_IS_TRUE_OR_FALSE(value) ((value) & (GR_TRUE | GR_FALSE))
#define GR_IS_PSEUDOTRUE(value) ((value)&GR_PSEUDOTRUE)



// [7] ADDITIONAL MACROS:
// (a) Remove the temporary value from a ground atom:
//      If the atom has both 'GR_TRUE' and 'GR_PSEUDOTRUE' flags set,
//      the atom should become 'GR_PSEUDOTRUE' after cleaning;
//      otherwise, atom should become GR_NORMAL.
#define GR_CLEAR_ATOM(atom) \
		{ \
		if((atom)->value&GR_TEMP) \
		  (atom)->value=(((atom)->value&(GR_TRUE|GR_PSEUDOTRUE)) == \
			(GR_TRUE|GR_PSEUDOTRUE))? GR_PSEUDOTRUE: GR_NORMAL; \
		}


// (b) Initialize limit of the current page:
// NOTE:  Last two elements in every page are reserved:
//       first: to allow input two elements at once without
//              checking for limit between the operations
//       second: to allow write DELIM_PAGE after the last element
//               in the current page
#define GET_LIMIT_FOR_PAGE(limit, page) \
	((limit)=(page)->data + (page)->size - 2)

// (c) Check whether the page will fit 'num' number of ground theory elements:
#define DOES_RULE_FIT_PAGE(num, page) ((num) < (page)->size - 2)



	// The ground theory is an array of such elements divided into pages:
	//  Depending on the particular element, the fields may have
	//  different meanings.
struct GroundTheoryElement
{
	union
	{
		AVLNode *atom;   // Atom or delimiter
		int num1;        // 1-st number
	};

	union
	{
		    // Reference to the next rule containing this ground atom:
		GroundTheoryElement *nextRule;
		int num2;        // 2-nd number
	};
};

	// The structure is needed to create a list of pages:
struct PageDescriptor
{
	GroundTheoryElement *data;  // Data in the page
	int size;                   // Size of memory allocated for data
	PageDescriptor *prev;       // Previous page
	PageDescriptor *next;       // Next page
};

struct GroundTheoryStatistics
{
		// Statistics from the ground rules
	unsigned int rules;        // Number of rules in the theory
	unsigned int cardAtoms;	   // Number of cardinality atoms in the theory
	unsigned int weightAtoms;  // Number of weight atoms in the theory

		// Statistics from the ground atoms.
	unsigned int atomsTrue;    // Number of true atoms in the theory
	unsigned int atomsFalse;   // Number of false atoms in the theory
		// Number of atoms OCCURRING IN THE GROUND THEORY,
		//  that have neither TRUE nor FALSE value
		//  (may be either NORMAL or PSEUDO-TRUE).
	unsigned int atomsUnknown;
	unsigned int atomsNodiff;
};


//  The purpose of this structure is provide consistent way of saving
//   information about the rule that is currently being input/copied.
//  This information is necessary if the current page doesn't have
//   enough space for the current rule and we need to move the rule
//    to the next page.
struct RuleInputData
{
		// Pointer to the next element to be written to
		//  (all the elements before are used by the rule)
	GroundTheoryElement *cur;
		// Current limit of the input. If we detect that cur >= limit,
		//   we need to go to the next page.
	GroundTheoryElement *limit;
		// Pointer to the starting element of the current rule.
	GroundTheoryElement *rule;
		// rule[nMeta] - points to the first element of the current
		//   meta-atom (this variable makes sense if we currently in
		//   the middle of some meta atom)
	int nMeta;
		// Current page:
	PageDescriptor *page;
};

class GroundRules
{
	//**************** DATA ***************************************
	Binary tree;   // AVL tree with ground atoms
		// References to the external objects:
	SymbolTable &symbt;
	StringConstants &strConst;

	StackL<AVLNode *> mainStk;      // Main propagation stack
		// Stack for lookahead to be able to clear the temporary values:
	StackL<AVLNode *> clearanceStk;

		// values for the variable state:
		// 0    -    Inside a list of simple atoms outside meta-atoms.
		// 1    -    Inside cardinality atom.
		// 2    -    Inside weight atom.
		// 3    -    The state before starting a meta-atom.
		// 4    -    Rule has not been started.
	int state;
	int isHeadI;//TRUE if the head is currently being input; FALSE otherwise

		// Main input structure
	RuleInputData input;

		// Static array of delimiters. "AVLNode *" pointers that point
		//  to the elements of this array are considered delimiters
		//  rather than regular atoms.
	static AVLNode delimiters[DELIM_NUM];
		// Text that explains the format of the human-readable output
		//  It will be written at the end of the human-readable file.
	static char humanOutputFormatExplanation[];

		// Variable to determine when to do the next compaction
	double tillNextCompaction;
		// Number of messages regarding repetitions in meta-lists:
	int numOfRepetitionMsg;

		// Variables to store information regarding the
		//  current rule that is being input:
	int isHorn; //TRUE if this is a horn atom; FALSE otherwise
	int curGroundRuleNum; // Current number of the ground rule
		// Line number of the source rule in the rule file:
	int curLineNum;

	int controlFlag; // EXECUTION CONTROL FLAG

		// Output streams:
	FILE *scriptOut;   // For script
	FILE *errorStream; // For errors about repetitions in meta-lists

		// Information regarding an inconsistency in the ground theory:
	GroundTheoryElement *conflictingRule;
	AVLNode *conflictingAtom;

	//**************** METHODS ***************************************

		// This method links one atom in a rule
	static void linkOneAtom(GroundTheoryElement *elementWithAtom,
	                        GroundTheoryElement *rule)
	{
			// If this is not the first occurrence of this atom
			//  in this rule, set 'nextRule' to NULL.
		if(elementWithAtom->atom->firstRule == rule)
			elementWithAtom->nextRule=0;
		else
		{
				// If this is THE FIRST occurrence,
				//  add the atom to the list:
			elementWithAtom->nextRule =
				elementWithAtom->atom->firstRule;
			elementWithAtom->atom->firstRule = rule;
		}
	}

	// [@2] GROUND ATOM ARGUMENT CONVERSION
	static unsigned int convertGroundAtomArguments(Atom *atom);
	static unsigned int atomNumFromTail(PPnode *pp, int dataArgIndex);
	void printAtom(FILE *file, int symbolno, unsigned int params);

	// [@3] OUTPUT METHODS:
	void printPageContents();
	void reportNotGroundedPredicates(PPnode *list);
	int writeAtomsToFile( FILE *file, unsigned int startingNumber,
	                      PPnode *list);
	void writeTheoryToFile( FILE *file, int isDIMACS,
	                        RuleNode *ruleSequence=0);
	void writeTheoryTo2Files( FILE *cnffile, FILE *pbfile);
	void writeAtomsToHRFile(FILE *file);
	void printNewAtomsInStack(GroundTheoryElement *rule, int oldSize);

	// [@4] PAGE LEVEL
	void calculateTheoryStatistics(GroundTheoryStatistics *stats);
	int impossibleCPLEX();
	void compactRules(int compactionFlag, GroundTheoryStatistics *stats=0);

	// [@5] MEMORY MANAGEMENT METHODS
	static PageDescriptor *pageAllocOrExpand(int minSize=0,
	                                         PageDescriptor *page=0);
	static void deletePage(PageDescriptor *page);
	static void moveToNextPage(RuleInputData *in, int needAlloc=1);
	void allocateAndInitializeFirstPage();
	void deleteAllPages();
	void clearTheory(GroundTheoryStatistics *stats);

	// [@6] HIGH LEVEL PROPAGATION
	unsigned int completePropagationAndAssignNumbers(
	    int isDIMACSfmt, int isLegacyFormat, GroundTheoryStatistics *stats);
	int propagateHornAtoms(int calculateNumbersOnly);
	int lookahead();

	// [@7] MIDDLE LEVEL PROPAGATION
	static void linkRule(GroundTheoryElement *rule);
	int finishPropagation(int isLookahead);
	int propagateRule(GroundTheoryElement *rule);
	void addSimpleRule(AVLNode *A, AVLNode *B);

	// [@8] LOW LEVEL PROPAGATION
	static GroundTheoryElement *clearOneRule(GroundTheoryElement *rule,
	                                         AVLNode *targetAtom);
	static GroundTheoryElement *propagateMetaAtom
		(StackL<AVLNode *> &mainAtomStk, GroundTheoryElement *&cur,
		AVLNode *targetAtom, int forcedValue);
	static GroundTheoryElement *propagateOneRule(
		StackL<AVLNode *> &mainAtomStk, GroundTheoryElement *rule,
		AVLNode *targetAtom, int isLookahead);
	static int getRuleNumber(GroundTheoryElement *rule);

	// Error handler for the INPUT INTERFACE [@9]:
	void inpError(char *method);
public:
	// [@1] MAIN METHOD
	int executeFinalStage(int executionFlag, PPnode *programPredicateList,
	         FILE *mainOutput, FILE *mainOutput2, FILE *humanOutput,
		 RuleNode *ruleSequence);

	// [@9] GROUND THEORY INPUT
	void startRule(int horn, int lineNum, int groundRuleNum);
	void addAtom(Atom *atom, int weight);
	void startMeta(int min, int max, int isWeight);
	void endAtomList();
	//void endMeta();
	void switchToHead();
	int endRule();

	// [@@] EXECUTION CONTROL INTERFACE:
		// Disable propagation, lookahead, and all compaction:
	void disableAnyPropagation()
	{
		controlFlag&=~(GR_DO_PROPAGATE | GR_DO_LOOKAHEAD |
			GR_DO_M_COMPACTION | GR_DO_ANY_COMPACTION);
	}
		// Disable one atom lookahead:
	void disableLookahead() {controlFlag&=~GR_DO_LOOKAHEAD;}
		// Disable compaction that is done over the whole theory
		//  time after time when a rule is being output.
	void disableMCompaction() {controlFlag&=~GR_DO_M_COMPACTION;}
		// Disable any compaction:
	void disableAllCompaction()
	{controlFlag&=~(GR_DO_M_COMPACTION | GR_DO_ANY_COMPACTION);}
		// Enable scripting to the specified output stream.
	void enableScript(FILE *scriptStream)
	{
		controlFlag|=GR_DO_SCRIPT; scriptOut=scriptStream;
		fprintf(scriptOut, "Human-readable output. For explanation of "
		"the format, see the end of the file.\n"
		"Writing atoms determined during propagation:\n");
	}

	// [@@] CONSTRUCTOR and DESTRUCTOR:
	GroundRules(SymbolTable &st, StringConstants &oc):
		symbt(st),strConst(oc), mainStk(10000, 2.0, 100000),
		clearanceStk(10000, 2.0, 100000)
	{
		state=4;     // Rule has not been started
		isHeadI=1;   // So that switching to head without starting rule

			// Initialize 'input' structure:
		allocateAndInitializeFirstPage();

		tillNextCompaction = GR_MIN_COMPACTION_SIZE;

		numOfRepetitionMsg=0;

		isHorn=0;
		curGroundRuleNum=0;
		curLineNum=-1;

			//By default, the propagation, compaction,
			//  and lookahead are enabled:
		controlFlag=GR_DO_PROPAGATE | GR_DO_LOOKAHEAD |
			GR_DO_M_COMPACTION | GR_DO_ANY_COMPACTION;

		errorStream=stderr;
		scriptOut=0;

		conflictingRule=0;
		conflictingAtom=0;
	}

	~GroundRules() { deleteAllPages();}
};



	// Error codes for the method 'Grounder::reportError'
#define GERR_NOERROR            0
#define GERR_DIVBYZERO          1
#define GERR_NOWEIGHT           2
#define GERR_BADWEIGHT          3

/*
class Grounder: the class that converts the predicate theory in the
intermediate representation into propositional ground theory that is
stored in the class 'GroundRules' by calling the methods in the
input interface
*/
class Grounder
{
		// Connection with the objects of the other classes:
	SymbolTable &symbt;
	GroundRules &out;
	StringConstants &strConst;

		// Evaluation stack: used by the method 'evaluate'
		//  to evaluate postfix expressions.
	EvalStack<int> stk;

		// Variables needed for error reporting:
	   // The code of the most recent error. This variable is used in
	   //  the method 'groundAtomList' to determine whether the
	   //  method 'evaluate' reported an error. In some cases,
	   // 'groundAtomList' has to report the error itself.
	int lastErrorCode;
	int errorCount;     // Grounding error count
	int lineNum;        // Line number of the rule currently being grounded
	FILE *errorStream;  // Stream to output all the error messages.

		// Function to report evaluation errors and negative weights:
	void reportError(int flag, Expression *exp, Atom *atom=0);

		// Evaluates any legal postfix arithmetic expression:
	int evaluate(Expression *exp);

		// Instantiates quantifiers: assigns values to the variables
		//  and checks conditions to be true:
	int instantiate(Quantifier *list, Quantifier * &current);

		// Methods used by the 'instantiate' to work
		//  with data predicate quantifiers:
		//  (1) Set the next value to the quantifier
	int setNextValue(Quantifier *quant);
		//  (2) Set the "value -1" to the quantifier so that subsequent
		//      call of 'setNextValue' will return the first value.
	void resetValue(Quantifier *quant);

		// Auxiliary methods for 'Grounder::run':
	void groundAtomList(Atom *list);
	void groundMetaAtomList(MetaAtom *mAtom, int isWeight);
public:
		// MAIN METHOD OF THE CLASS. Runs everything else:
	int run(RuleNode *theory);

		// Constructor:
	Grounder::Grounder(SymbolTable &st, GroundRules &groundTheory,
		StringConstants &oc): symbt(st), out(groundTheory),
		strConst(oc), stk(100)
	{
		lastErrorCode=GERR_NOERROR;
		errorCount=0;
		lineNum=-1;
		errorStream=stderr;
	}
};
