/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: common.cpp
 *
 *  CONTENTS:   This file contains the implementation of the classes:
 *              'StringConstants', 'TraverseVariables', 'TraverseAtoms'
 *
 *  DATE:     March 19, 2004
 ***************************************************************************/

#include <stdarg.h>
#include "common.h"

////*****************************************************************
/////////////  STRING CONSTANTS  ////////////////////////////////////
//******************************************************************

	// Constructor: Initializes everything to zero
StringConstants::StringConstants(SymbolTable &st):symbt(st)
{
	constBase=finalFlag=0;
	constList=-1;
	minNumber=0;
	constIndex=0;
	constNum=0;
}

	// Destructor: Deallocates all memory
StringConstants::~StringConstants()
{
	if(constIndex) free(constIndex);
	constIndex=0;
}


/*
char *StringConstants::getConstantFromNumber(int number)
DESCRIPTION
	Given a number corresponding to a constant (NOT symbol number),
	returns the name of that constant
*/
char *StringConstants::getConstantFromNumber(int number)
{
	if(!finalFlag) error("StringConstants::getConstantFromNumber");
	number-=constBase;
	if(number < 0 || number >= constNum) return 0;
	return (symbt.getName(constIndex[number]));
}


/*
int getNumberForConstant(int symbolno)
DESCRIPTION
	Given a symbol number of a constant, returns the number assigned 
	to this constant.
*/
int StringConstants::getNumberForConstant(int symbolno)
{
	if(!finalFlag) error("StringConstants::getNumberForConstant");
	return((symbt.getVal(symbolno))->number);
}

/*
void StringConstants::printConstant(FILE *out, int number)
DESCRIPTION
	Prints the constant corresponding to a given number (NOT symbol number)
	to the given stream

PARAMETERS
	out     - output stream to print the constant to
	number  - number for constant
*/
void StringConstants::printConstant(FILE *out, int number)
{
	char *str;
	if((str=getConstantFromNumber(number)))
		fprintf(out, "%s", str);
	else
		fprintf(out, "%d", number);
}

/*	
void StringConstants::getConstantRange(int *min, int *max)
DESCRIPTION
	Obtains the range of numbers that are treated as string constants
	if (*min) <= i <= (*max), then i is treated like a constant
PARAMETERS
	min	 - pointer to a variable to receive minimum (may be NULL)
	max  - pointer to a variable to receive maximim (may be NULL)
*/
void StringConstants::getConstantRange(int *min, int *max)
{
	if(min) *min=constBase;
	if(max) *max=constBase+constNum-1;
}

/*
The following class defines an abstract array of string constants
It is based on 'consIndex' array, but comparison is defined as 
lexicographic comparison of the corresponding constants.
*/
class StringConstantArray
{
	SymbolTable &st;	// Need a symbol table in here.
	int *index;			// Constant index
	int size;			// Number of constants
public:
		// Needed to initialize an object
	void insert(int *constIndex, int constNum)
	{index=constIndex; size=constNum;}

		// Swap two elements of an array: just simple swapping of
		//  elements in the constant index
	void swap(int i, int j)
	{ int temp=index[i]; index[i]=index[j]; index[j]=temp; }

		// Comparison by comparing names using strcmp:
	int isLess(int i, int j)
	{ return (strcmp(st.getName(index[i]), st.getName(index[j])) < 0); }

	int getNumOfElements() {return size;}
	StringConstantArray(SymbolTable &symbt):st(symbt) {size=0;}
};

/*
void StringConstants::addConstant(int symbolno)
DESCRIPION
	This method should be called whenever the new constant appears in the
	program (is allowed to be called for any constant). It initialized all
	necessary data fields and adds the constant to a temporary list of 
	string constants (to be used later by buildReferenceStructure)

PARAMETERS
	symbolno - symbol number of a constant
*/	
void StringConstants::addConstant(int symbolno)
{
	STVal *constVal;

	if(symbt.getType(symbolno)==TYPE_NONE)
	{
		symbt.setType(symbolno, TYPE_STRING_CONST);
		constVal=symbt.getVal(symbolno);
		constVal->number=-1;
		
			// Creating temporary list of constants.
		constVal->number=constList;
		constList=symbolno;
		constNum++;
	}
}


/*
int StringConstants::buildReferenceStructure(int *minNumberPtr,int *constNumPtr)
DESCRIPTION
	This method assigns numbers to constants (saved in the symbol table),
	allocates and initializes array 'consIndex', so that a constant name
	can be retrieved based on a number and vice versa.
	It may return values of the minimum number from dataSequence and
	number of constants in case of error

PARAMETERS
	minNumPtr   - Variable receives the miminal number from
	                the declaration sequence (may be NULL)
	constNumPtr - Variable receives the number of constants
	                in the program (may be NULL)

RETURNS
	Either SE_NO_ERROR if everything is OK, or SE_PROBLEM_CONST if there
	are not enough numbers for all string constants.
*/
int StringConstants::buildReferenceStructure(int *minNumberPtr,int *constNumPtr)
{
	STVal *stVal;
	int i,retValue=SE_NO_ERROR;
	StringConstantArray sortArray(symbt);

	if(finalFlag) error("StringConstants::finalize");
	finalFlag=1;

		//  ASSIGNING NUMBERS TO THE CONSTANTS:

		// Check whether we have enough numbers to assign
	if(INTEGER_MIN+constNum>minNumber)
	{	// We don't
		// Assign the numbers anyway, but set the error values:
		if(minNumberPtr) *minNumberPtr=minNumber;
		if(constNumPtr) *constNumPtr=constNum;
		retValue=SE_PROBLEM_CONST;
		constBase=INTEGER_MIN;
	}
	else
		constBase=minNumber-constNum;

		// Allocate the memory for the constant indexing array:
	constIndex=(int *)malloc(constNum*sizeof(int));

	if(constNum > 0 && !constIndex)
	{
		fprintf(stderr, "CRITICAL ERROR in StringConstants::"
			"buildReferenceStructure\n"
			"Cannot allocate memory\n\n");
		exit(1);
	}

		// Save the symbol numbers of all the constants in the array:
	for(i=0;constList!=-1&&i<constNum;i++)
	{
			// Save the constant number in the indexing array
		constIndex[i]=constList;

			// Retrieve the number of the next constant:
		stVal=symbt.getVal(constList);
		constList=stVal->number;
	}

	if(constList!=-1||i!=constNum)	// Checking for errors
	{
		fprintf(stderr, "CRITICAL ERROR: The list of constants "
			"has been built incorrectly\n");
		exit(1);
	}

		// Sort the reference array to arrange constants alphabetically:
	sortArray.insert(constIndex, constNum);
	quickSort<StringConstantArray>(sortArray);

		// Assign the number for each constant to the value in the
		//  symbol table.
	for(i=0;i<constNum;i++)
		symbt.getVal(constIndex[i])->number = constBase + i;

	return(retValue);
}

	// Internal critical error message function:
void StringConstants::error(char *nameOfMethod)
{
	fprintf(stderr, "CRITICAL ERROR in method %s\n"
		"The method cannot be called at the given moment\n\n",
		nameOfMethod);
	exit(1);
}


////*****************************************************************
/////////////  TRAVERSE ATOMS  ////////////////////////////////////
//******************************************************************

// Constructor: initializes variables.
TraverseVariables::TraverseVariables()
{
	level=levels-1;
	startFlag=0;
}

/*
void TraverseVariables::start(char *controlString, ...)
DESCRIPTION
	Start traversing the meta list

PARAMETERS
	a) For user's calls: controlString == "<L1><L2>...<LN>",
		where <Li> - is a letter which corresponds to the type
		of a list to be traversed:
			A  -  List of atoms from the body
			H  -  List of atoms from the head
			E  -  List of expressions
			Q  -  Lits of quantifiers
		Then control string should be followed by N pointers
		to the corresponding lists.

	b) For calls from the 'next': controlString == "<L1><L2>...<LN>*"
		<Li> are the same. However, it should be followed by pairs:
		pointer, atomType.

EXAMPLE:
User's call: start("AHE", listOfAtoms1, listOfAtoms2, listOfExpressions).
Call from 'next': start("E*", listOfExpressions, TRAV_IN_DATAPOS).
*/
void TraverseVariables::start(char *controlString, ...)
{
	va_list args;
	Parts *part;
	int i;
	int curPart;


	if(!controlString) return; // Check for zero pointer:

		// Check for empty string:
	curPart=strlen(controlString);
	if(curPart<=0) return;

	va_start(args, controlString);	// Start reading the arguments:

		// If this was user's call, need to start from the first level
		//  Otherwise, go to the next level.
	if(controlString[curPart-1]=='*')
	{
		curPart--; // Go to the valid letter.
		level++;   // Go to the next level
	}
	else
		level=levels; // Start from first level

		// Check the conditions to prevent accesses out of the
		//  array boundaries:
	if(level-levels>=TRAV_MAX_LEVELS-1)
	{
		fprintf(stderr, "CRITICAL ERROR: in the method "
			"TraverseVariables::start\n"
			"too many nested structures:"
			" error during their building\n");
			exit(1);
	}
	if(curPart>TRAV_MAX_PARTS)
	{
		fprintf(stderr,"CRITICAL ERROR: in the method "
			"TraverseVariables::start\n"
			"the control string(first parameter) is too long\n\n");
		exit(1);
	}

		// Analyze the control string and write the corresponding parts:
	for(i=0,part=level->parts+curPart-1;i<curPart;i++,part--)
	{
		part->atomType=0;

			// Read the corresponding pointer and set
			//  the corresponding token type:
		switch(controlString[i])
		{
		case 'E':
			part->exp=va_arg(args, Expression *);
			part->token=TRAV_EXPR;
			break;
		case 'H':
			part->atomType=TRAV_IN_HEAD;
		case 'A':
			part->atom=va_arg(args,  TempAtom *);
			part->token=TRAV_ATOM;
			break;
		case 'Q':
			part->quant=va_arg(args,  Quantifier *);
			part->token=TRAV_QUANT;
			break;
		default:
			fprintf(stderr, "CRITICAL ERROR: in the method "
				"TraverseVariables::start\n"
				"wrong character \'%c\' in the control string."
				"\n\n",controlString[i]);
			va_end(args);
			exit(1);
		}

			// If this was not user's call, calculate the
			// resulting atom type: take the type from the
			// previous level and add the type given a parameter:
		if(level>levels)
			part->atomType=(level-1)->part->atomType|va_arg(args, int);
	}

		// Start analyzing parts from the last.
	level->part=level->parts+curPart-1;
	startFlag=1;	// Indicate that 'start' was called.

	va_end(args);	// End arguments
}


/*
VarNode *TraverseVariables::next(int *atomType, int *paramType)
DESCRIPTION
	Returns the next variable node in the traversal sequence

PARAMETERS
~~~~ NOT USED ANYWHERE AND THEREFORE NOT LISTED HERE ~~~~~~~~
*/
// The following macro gets the next pointer of any type and if next
//  pointer is NULL, it goes to the previous level and breaks from the
//  'switch' statement where it should be put to.
#define TRAV_NEXT_ITEM(KIND) \
	if(level->part->KIND==0 || startFlag==0 &&      \
	(level->part->KIND=level->part->KIND->next)==0) \
	{ level->part--; startFlag=1;break;}            \
	KIND=level->part->KIND; startFlag=0;

VarNode *TraverseVariables::next(int *atomType, int *paramType)
{
    TempAtom *atom;
    Expression *exp;
    Quantifier *quant;

        // Loop over levels until we are done with the highest level:
    for(;level>=levels;level--,startFlag=0)
    {
            // Loop over all parts until we are done with the first part:
        while(level->part>=level->parts)
        {
                // Switch over type of a part
                //  (i.e. EXPRESSION, ATOM, or QUANTIFIER)
            switch(level->part->token)
            {
            case TRAV_EXPR:
                   // Go to the next expression if necessary:
                TRAV_NEXT_ITEM(exp);

                   // Proceed according to the expression type:
                switch(exp->token)
                {
                    // If there is another expession
                    // attached to the node, traverse it.
                case EXPR_EXPR:
                    start("E*",exp->exp, 0);
                    break;

                    // If this expression node contains
                    //  a variable, return this variable:
                case EXPR_VAR:
                case EXPR_VAR_QUANT:
                case EXPR_ASSIGN:
                       // Return the atomType and paramType:
                    if(atomType) *atomType=level->part->atomType;
                    if(paramType) *paramType=(level>levels &&
                        (level-1)->part->token==TRAV_EXPR &&
                            (level-1)->part->exp->token==EXPR_EXPR)?
                                 EXPR_EXPR:exp->token;

                    return(exp->var);

                    // If this is data predicate condition,
                    //  traverse the argument list of the
                    //  data predicate:
                case EXPR_DPOSITIVE:
                    start("E*", exp->atom->arg, TRAV_IN_DATAPOS);
                    break;
                case EXPR_DNEGATIVE:
                    start("E*", exp->atom->arg, TRAV_IN_DATANEG);
                    break;
                }
                break;

            case TRAV_ATOM:
                // Go to the next atom if necessary:
                TRAV_NEXT_ITEM(atom);

                switch(atom->type)
                {
                    // Program atom: traverse arguments, weight expression,
                    // local variables, and data restrictions
                case ATOM_PROG:
                    start("EEQA*", atom->prog.arg, TRAV_IN_PREDICATE,
                          atom->prog.weight,       TRAV_IN_WEIGHTASS,
                          atom->prog.quant,        TRAV_IN_QUANT,
                          atom->prog.restrictions, TRAV_IN_RESTRICTIONS);
                    break;

                    // Data predicate: travese arguments
                case ATOM_DATA:
                    start("E*", atom->data.arg, TRAV_IN_DATAPOS);
                    break;
                case ATOM_DATANEG:
                    start("E*", atom->data.arg, TRAV_IN_DATANEG);
                    break;

                    // Predefined atom: traverse the expression:
                case ATOM_CONDITION:
                    start("E*", atom->cond.first, TRAV_IN_PREDEFINED);
                    break;

                    // Cardinality or weigh atom: travese the
                    // minimum, atom list, maximum.
                case ATOM_CARD:
                    start("EAE*", atom->meta.min,  0,
                                  atom->meta.list, TRAV_IN_CARD,
                                  atom->meta.max,  0);
                    break;
                case ATOM_WEIGHT:
                    start("EAE*", atom->meta.min,  0,
                                  atom->meta.list, TRAV_IN_WEIGHT,
                                  atom->meta.max,  0);
                    break;
                }
                break;

            case TRAV_QUANT:
                // Go to the next quantifier if necessary:
                TRAV_NEXT_ITEM(quant);

                    // If this is a variable quantifier,
                    //  start traversing condition, but return
                    //  the variable.
                if(quant->token==QUANT_VAR)
                {
                    start("E*",quant->conditions,  TRAV_IN_QUANT_COND);

                    if(atomType) *atomType=level->part->atomType;
                    if(paramType) *paramType=EXPR_VAR;

                    return(quant->var);
                }
                else    // Traverse conditions and data predicate arguments:
                    start("EE*",quant->conditions,  TRAV_IN_QUANT_COND,
                                quant->arg,         TRAV_IN_QUANT_VAR);
                break;
            }
        }
    }

    return 0;
}





////*****************************************************************
/////////////  TRAVERSE ATOMS  ////////////////////////////////////
//******************************************************************

// Constructor:
TraverseAtoms::TraverseAtoms()
{
		// Initialize variables:
	cur1=cur2=0;
	curPart=0;
	curSurroundType=0;
	traverseMetaAtoms=0;
	memset(&dummyAtom, 0, sizeof(dummyAtom));
}

/*
void TraverseAtoms::start(TempAtom *body, TempAtom *head, int travMetaAtoms)
DESCRIPTION
	Start traversal of the rule. Can traverse body only, head only, or the
	whole rule.

PARAMETERS
	body          -  List of atoms in the body of a rule (may be NULL)
	head          -  List of atoms in the head of a rule (may be NULL)
	travMetaAtoms -  When this flag is set, the TraverseAtoms::next
	             will return the meta atoms before proceeding to meta-lists
*/
void TraverseAtoms::start(TempAtom *body, TempAtom *head, int travMetaAtoms)
{
	cur1=cur2=0;
	curSurroundType=0;
	parts[0]=head;
	parts[1]=body;
	curPart=2;	// Start from the last.
	traverseMetaAtoms=(travMetaAtoms)?1:0;
}


/*
TempAtom *TraverseAtoms::next(int *surroundType)
DESCRIPTION
	Retruns the next atom in the traversal sequence.

PARAMETERS
	surroundType - Pointer to a variable that can receive the value
	               indicating whether the atom is in head and/or meta-list
*/
TempAtom *TraverseAtoms::next(int *surroundType)
{
		// If we are on the level of meta-lists, proceed to the
		// next atom on this level:
	if(cur2)
	{
		cur2=cur2->next;
			// Set the 'surroundType'
		if(surroundType) *surroundType=curSurroundType;

		     // If we reached the end of list, then we are no more in
		     //  cardinality or weight atom => adjsut 'curSurroundType'
		     // If we didn't reach the end, return the current element.
		if(cur2) return(cur2);
		else
			curSurroundType&=~(TRAV_IN_WEIGHT|TRAV_IN_CARD);
	}

		// Go to the next element:
	if(cur1) cur1=cur1->next;

		// If the next element is zero, scan the parts until we find
		//  non-zero element, or there will be no more parts, in which
		//  case we return 0.
	while(!cur1)
	{
		if(curPart>0)
		{
			curPart--;
			cur1=parts[curPart];

				// parts[0] - stores pointer to the head.
			if(curPart==0) curSurroundType|=TRAV_IN_HEAD;
		}
		else
			return 0;
	}

		// Check whether the current atom is a meta atom and
		//  if so adjust 'curSurroundType'.
	if(cur1->type==ATOM_CARD) curSurroundType|=TRAV_IN_CARD;
	if(cur1->type==ATOM_WEIGHT) curSurroundType|=TRAV_IN_WEIGHT;

		// If we are dealing with meta-atom:
	if(cur1->type==ATOM_CARD||cur1->type==ATOM_WEIGHT)
	{
			// Append the meta-list to the dummy atom and set
			//  'cur2' to this atom so that next
			//  call to the 'next' will return the first atom
			//  in that list
		dummyAtom.next=cur1->meta.list;
		cur2=&dummyAtom;

			// If we need to return meta atom before proceeding to
			//  the meta-list, do it.
		if(traverseMetaAtoms)
		{
			if(surroundType) *surroundType=curSurroundType&
				(~(TRAV_IN_CARD|TRAV_IN_WEIGHT));
			return(cur1);
		}

			// Otherwise, proceed to the meta list:
		return(next(surroundType));
	}

		// 'cur1' points to a simple atom. We need to return it.
	if(surroundType) *surroundType=curSurroundType;
	return(cur1);
}
