/*==========================================================================;
 *
 *  FILE NAME: grammar.y
 *
 *  CONTENTS:  Bison Grammar for PSGRND.
 *
 *  DATE:     March 24, 2004
 ***************************************************************************/

/*******************************************************************/
/*****************      DECLARATIONS      **************************/
/*******************************************************************/
/* Necessary lines in the yacc output file: */
%{
#include "rules.h"

    /* Define the instance of the class "Rules" the parser will work with.*/
#define P rules

#define YYERROR_VERBOSE

    /* Input and error functions of yacc.*/
extern int yylex();
extern void yyerror(char const *mes);

    /* External objects for the yacc: */
extern Rules rules;
extern SymbolTable st;
extern LexicalAnalyzer la;
%}
    /* Starting nonterminal */
%start program

    /* Union of all possible types of the nonterminals */
%union
{
   int number;
   ExpressionHandler *expressionHandler;
   Value *value;
   VListHandler *vlist;
   TempList *tlist;
   VarNode *var;
   TempAtom *atom;
   Expression *expression;
   Quantifier *quant;
}


/* [1] Types of the terminals: */
%token <number> NUM ID DataPredID WeightFunID
%token <number> ProgramPredID VariableID ConstantID

%token PRED VAR ABS MOD MIN MAX
%token ARROW COLONMINUS DOUBLEDOT EQUAL NOTEQUAL NOTGREATER NOTLESS

/* [2] Types of the nonterminals: */

    /* (a) Data atom and weight function definitions */
%type <value>  constant
%type <vlist> constantList dataRange

    /* (b) Variable declaration */
%type <tlist> dataIdList
%type <var> variableList

    /* (c) Expression processing */
%type <number> addop mulop relop unaryId binaryId
%type <expressionHandler> factor additiveExpression term
%type <expressionHandler> unaryFunction binaryFunction weightFunction

    /* (d) Level of argument lists of atoms */
%type <expressionHandler> argument argumentList moreArgumentList operand
%type <expressionHandler> existentialList moreExistentialList

    /* (e) Level of atoms in the rules */
%type <expressionHandler> value
%type <quant> localVariables localVarList
%type <atom> atom simpleProgramAtom programAtom restrictionAtom
%type <atom> cardinalityAtom weightAtom predefinedAtom dataAtom
%type <atom> cardinalityList weightList weightAssignment
%type <atom> dataRestrictions
%type <atom> antecedent antecedentList consequent consequentList
/*
   ACTION NOTATION:
1) ...New   -	The action creates a new object and returns a pointer to it
2) ...Add   -	The action (possibly) creates a new object and inserts it
               into a list.
               First Parameter:  list.
               Second Parameter: object.
3) Other actions are designed for special purposes.

   ABBREVIATION IN THE NAMES OF ACTIONS:
1) exp  - Expression
2) var  - Variable
3) Card - Cardinality
 */

%%
/*******************************************************************/
/*****************      RULES      *********************************/
/*******************************************************************/

/*              Grammar for PS + */
program : dataSequence          {  P.rebuild(); }
          declarationSequence   { if(!P.checkErrorsBeforeRules()) YYABORT; }
          ruleSequence          { if(!P.finishParse()) YYABORT; };



/*              Data Sequence   */
dataSequence : /* empty */
             | dataSequence dataDefinition
             | dataSequence error '.' {yyerrok;};

dataDefinition : dataAtomDefinition | weightFunctionDefinition ;

dataAtomDefinition :         ID '(' constantList ')' '.'{P.dataNew($1, $3, 0);}
                   |         ID '(' dataRange    ')' '.'{P.dataNew($1, $3, 0);}
                   | DataPredID '(' constantList ')' '.'{P.dataNew($1, $3, 0);}
                   | DataPredID '(' dataRange    ')' '.'{P.dataNew($1, $3, 0);};

weightFunctionDefinition : ID '(' constantList ')' '=' additiveExpression
                           /*TYPE:NUM*/ '.' { P.dataNew($1, $3, $6);}
                      | WeightFunID '(' constantList ')' '=' additiveExpression
                           /*TYPE:NUM*/ '.' { P.dataNew($1, $3, $6);};

dataRange : additiveExpression /*TYPE: NUM*/ DOUBLEDOT additiveExpression
                /*TYPE: NUM*/ {$$=P.valueNewRange($1, $3);};

constantList : constant                  {$$=P.valueAdd(0, $1); }
             | constantList ',' constant {$$=P.valueAdd($1, $3);};

constant : ConstantID    {$$=P.valueNew(EXPR_STRING, $1      );}
         | NUM           {$$=P.valueNew(EXPR_NUM,    $1      );}
         | '-' NUM       {$$=P.valueNew(EXPR_NUM,    (-($2)) );}
         | ID            {$$=P.valueNew(EXPR_STRING, $1      );};



/*              Program Declaration Sequence */
declarationSequence : declaration | declarationSequence declaration ;

declaration : programAtomDeclaration | variableDeclaration ;

programAtomDeclaration
 : PRED ID '(' dataIdList ')' ':' DataPredID '.' {P.programPredNew($2, $4, $7);}
 | PRED ID '(' dataIdList ')' '.'                {P.programPredNew($2, $4);}
 | PRED ID '.'                                   {P.programPredNew($2);}
 | PRED error '.' {yyerrok;};

variableDeclaration : VAR DataPredID variableList '.'
                                    {P.varDeclarationNew($2, $3);}
                    | VAR error '.' {yyerrok;};

dataIdList : DataPredID                 {$$=P.dataIdAdd(0, $1);}
           | dataIdList ',' DataPredID  {$$=P.dataIdAdd($1, $3);};

variableList : ID                  {$$=P.varAdd(0, $1);}
             | variableList ',' ID {$$=P.varAdd($1, $3);};



/*              Rule Sequence*/
ruleSequence : /* empty */ | ruleSequence rule
             | ruleSequence error '.' {yyerrok;};

rule : hornRule | implication ;

hornRule : simpleProgramAtom COLONMINUS antecedent '.' {P.ruleNew($3,$1,1);};

implication : antecedent ARROW consequent '.' {P.ruleNew($1,$3,0);}
            | consequent '.'                  {P.ruleNew(0,$1,0);};

antecedent : /* empty */ {$$=0;} | antecedentList ;

antecedentList : atom | antecedentList ',' atom {ATOM_ADD($$, $1, $3)};

consequent : /* empty */ {$$=0;} | consequentList ;

consequentList : atom | consequentList '|' atom {ATOM_ADD($$, $1, $3)};


/*              Atom, Predefined Atom, Data Atom */
atom : restrictionAtom | programAtom | cardinalityAtom | weightAtom ;

restrictionAtom: predefinedAtom | dataAtom;

predefinedAtom : operand /*TYPE:any, except WEIGHT */ relop
                 operand /*TYPE:any, except WEIGHT */
                           {$$=P.predAtomNew($1, $3, $2);};

operand  : additiveExpression | ConstantID {$$=P.expNew($1, EXPR_STRING);};

dataAtom : DataPredID '(' argumentList ')' {$$=P.dataAtomNew($1, $3);};


/*              Simple Atom */
simpleProgramAtom : ProgramPredID '(' argumentList ')'
                                  {$$=P.programAtomNew($1, $3, 0, 0);}
                  | ProgramPredID {$$=P.programAtomNew($1, 0,  0, 0);};

argumentList : moreArgumentList argument {$$=P.argAdd($1, $2);};

moreArgumentList : moreArgumentList argument ',' {$$=P.argAdd($1, $2);}
                 | /* empty */ {$$=0;};

argument : ConstantID {$$=P.expNew($1, EXPR_STRING);}
         | additiveExpression/*TYPE:NUM,VAR,NUM+VAR,STRVAR,MIXVAR*/;


/*              Existential Atom */
programAtom
      : ProgramPredID '(' existentialList ')' localVariables dataRestrictions
                        {$$=P.programAtomNew($1, $3, $5, $6);}
      |           simpleProgramAtom           localVariables dataRestrictions
                        {$$=P.programAtomFromSimple($1, $2, $3);};

existentialList : moreArgumentList '_' moreExistentialList
                                            {$$=P.argUndMerge($1,$3);};

moreExistentialList : moreExistentialList ',' argument {$$=P.argAdd($1, $3);}
                    | moreExistentialList ',' '_'      {$$=P.argAdd($1,  0);}
                    | /* empty */                      {$$=0;};

dataRestrictions : /* empty */                          {$$=0;}
                 | dataRestrictions ':' restrictionAtom {ATOM_ADD($$, $1, $3)};

localVariables : /* empty */ {$$=0;} | '[' localVarList ']' {$$=$2;};

localVarList : VariableID                  {$$=P.localVarAdd(0,  $1);}
             | localVarList ',' VariableID {$$=P.localVarAdd($1, $3);};


/*              Cardinality Atom */
cardinalityAtom : value/*TYPE:VAR,NUM,+*/ '{' cardinalityList '}' value
                  /*TYPE:VAR,NUM,+*/ {$$=P.metaAtomNew($3, $1, $5, ATOM_CARD);};

cardinalityList : programAtom
                | cardinalityList ';' programAtom {ATOM_ADD($$, $1, $3)};

value : /* empty */ {$$=0;} | additiveExpression;


/*              Weight Atom */
weightAtom : value /*TYPE:NUM,VAR,+*/'{' weightList '}' value/*TYPE:NUM,VAR,+*/
             {$$=P.metaAtomNew($3,$1,$5, ATOM_WEIGHT);};

weightList : weightAssignment
           | weightList ';' weightAssignment {ATOM_ADD($$, $1, $3)};

weightAssignment : simpleProgramAtom '=' additiveExpression/*TYPE: WEIGHT,NUM,+*/
                   localVariables dataRestrictions
                   {$$=P.weightAssignmentFromSimple($1, $3, $4, $5);};

/*              Expressions */
additiveExpression : term | '-' term     {$$=P.expMerge($2, 0, EXPR_UMINUS);}
                   | additiveExpression addop term {$$=P.expMerge($1,$3,$2);};


term : factor | term mulop factor {$$=P.expMerge($1,$3,$2);};

factor : '(' additiveExpression ')'  {$$=$2;}
       | binaryFunction
       | unaryFunction
       | weightFunction
       | VariableID                  {$$=P.expNew($1, EXPR_VAR);}
       | NUM                         {$$=P.expNew($1, EXPR_NUM);}
       | ID     /* Error rule */     {$$=P.expErrorID($1);};


/*              Functions */
unaryFunction : unaryId '(' additiveExpression ')' {$$=P.expMerge($3,0,$1);};

binaryFunction : binaryId '(' additiveExpression ',' additiveExpression ')'
                 { $$=P.expMerge($3, $5, $1); };

weightFunction : WeightFunID '(' argumentList ')' {$$=P.weightCallNew($1,$3);};

unaryId :  ABS {$$=EXPR_ABS;};
binaryId : MOD {$$=EXPR_MOD;} | MIN {$$=EXPR_MIN;} | MAX {$$=EXPR_MAX;};


/*              Relational Operators */
relop : '<'   {$$=EXPR_LESS;}    | NOTGREATER {$$=EXPR_NOTGREATER;}
      | '>'   {$$=EXPR_GREATER;} | NOTLESS    {$$=EXPR_NOTLESS;}
      | EQUAL {$$=EXPR_EQUAL;}   | NOTEQUAL   {$$=EXPR_NOTEQUAL;};


/*              Arithmetic Operators    */
addop : '+' {$$=EXPR_PLUS;} | '-'{$$=EXPR_MINUS;};
mulop : '*' {$$=EXPR_MUL; } | '/'{$$=EXPR_DIV;};

%%
/*******************************************************************/
/*****************      SUBROUTINES       **************************/
/*******************************************************************/

/* auxiliary procedures for the parser: */

/* The following function determines whether parser did one token lookahead */
int isParserLookahead() {return (yychar != YYEMPTY); }

/* Standard bison routine to print SYNTAX ERROR MESSAGES */
void yyerror(char const *message)
{
   int number, token;

      /* If the input has been interrputed, ignore all parsing errors: */
   if(la.isInputInterrupted()) return;

      /* Print the name of the current file: */
   la.reportCurrentFileName(P.getErrorStream());

      /* Print the line number, token name and the error message: */
   token = la.getLastToken(&number);
   fprintf(P.getErrorStream(), "ERROR(line %d), token \'%s\': %s\n",
           la.getLineNum(0),la.tokenToString(token, number), message);

      /* Report the syntax error to the parser: */
   P.registerSyntaxError();
}

/* Standard bison routine to return next token to the parser */
#if (DEBUG_PRINT_TOKENS)
int yylex()
{
        int num=la.scan(&(yylval.number));

        printf("%s   %d\n", la.tokenToString(num, yylval.number),
                la.getLineNum(0));
        return num;
}
#else
int yylex() {return la.scan(&(yylval.number));}
#endif
