/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: memory.h
 *
 *  CONTENTS: Classes for memory management. Can be used as intermediate
 *              interface between the user program and the memory allocation
 *              routines such as malloc and free to allocate the memory
 *              more efficiently. Classes:
 *              MemAlloc, MemNode, MemNodeN.
 *
 *  OPTIONS:  MEMORY_H_DEBUG (see definition)
 *
 *  DATE:     February 24, 2004
 *  VERSION:  1.7
 ***************************************************************************/

#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>

// Preventing redefinitions if this file is included more than one time.
#ifndef _INC_MEMORY_MANAGER
#define _INC_MEMORY_MANAGER


// If the following is uncommented, the file will be compiled in the debug mode
//  This enables method 'printStatistics'.
//#define MEMORY_H_DEBUG


// Default parameters for the constructors
	//   Size of the initial piece for 'MemAlloc'
#define MEM_DEF_SIZE	100
	//  Maximal size of the piece for 'MemAlloc'
#define MEM_DEF_MAX	1000000
	// Default growth factor for 'MemAlloc' and 'BaseMemNode'
#define MEM_DEF_FACTOR	2.0
	// Default starting number of nodes for 'BaseMemNode'
#define NODE_DEF_NUM	24
	// Default Maximal Number of nodes for 'BaseMemNode'
#define NODE_DEF_MAX	240000

// Sizes of pieces(pages) of memory allocated internally by these classes will
//  be divisible by ROUNDOFF.
#define ROUNDOFF 8


// Linked list of these structures contains the information about allocated
//  pages of memory. All the classes don't allocate nodes of this structure
//  separately, but rather make it a part of the allocated page
//  (correspondingly increasing its size).
struct MemPageDescriptor
{
	char *page;                // Pointer to the corresponding page
	MemPageDescriptor *next;   // Next descriptor.
};



/*
Class MemAlloc:

Provides the fast routines for allocating memory of arbitrary size
 (works best with small-medium sizes: 1 .. 10000 bytes,).
 However, one piece of memory allocated using this class cannot be
 released for future reuse, other than by releasing ALL memory
 allocated by a specific object of a given class.
 In contrast with malloc, the class is guaranteed to return a
 pointer to the piece of memory with the requested size

 The class allocates BIG pieces (later called pages) of memory using the
 method 'malloc' and then redistributes their parts to the user program.

 BIG pieces are allocated only when demanded.

 The use of this class may cause internal fragmentation, which is
 negligible if good parameters are given to the constructor.

MAIN METHOD: alloc(size).
*/
class MemAlloc
{
	MemPageDescriptor *pageList; // Linked list of internal pages allocated
	                             //  by an object of the class using malloc

	size_t nextSize; // Size of the next page to be allocated.
	char *curPtr;    // Pointer to the first free byte in the current page
	char *maxPtr;    // Pointer, which points to the byte right after
	                 //  the last byte of the curent page

	const size_t FIRST_PAGE_SIZE; // Size of first page to be allocated
	const float GROWTH_RATIO; // Growth factor: the size of next page to be
	                          // allocated is GROWTH_RATIO as big as the
	                          //  size of the current page.
	const size_t MAX_PAGE_SIZE; // Maximal size of the page to be allocated
		// Variables for debug purposes:
	#ifdef MEMORY_H_DEBUG
		int memFree;    // Remaining free memory in the last buffer
		int memBusy;    // Memory allocated by the user.
		int total;      // Total memory allocated for pages
		                //  i.e. the total number of bytes allocated
				//    by the methods of the class
		int reallocNum; // The number of pages that have been allocated
		int restartNum; // The number of calls of the method 'freeAll'.
	#endif

public:
	/*
	void *alloc(size_t size)
		The function allocates a piece of memory. In most cases it
		will be a part of the current page.

	PARAMETERS, RETURN: The same as for standard function 'malloc'.
	*/
	void *alloc(size_t size)
	{
		int actualSize;

		    // Trying to return a piece of the current page.
		void *temp=curPtr;
		if(size<0)
		{
			fprintf(stderr, "CRITICAL ERROR in the method "
				"\'MemAlloc::alloc\'\n"
				"size of memory to allocate is %d: should "
				"NOT be negative\n",size);
			exit(1);
		}
		if((curPtr+=size)<=maxPtr)
		{
			#ifdef MEMORY_H_DEBUG
				memBusy+=size;
				memFree-=size;
			#endif
			return(temp);
		}

		// NOT ENOUGH MEMORY IN THE CURRENT PAGE: NEED TO ALLOCATE
		//  additional page and return a part of it.

		    // Calculating the size of the page to allocate.
		    // By default it is nextSize, but it should not be less
		    //  then the requested allocation size and should not
		    //  be greater than MAX_PAGE_SIZE.
		    // It also should be divisible by ROUNDOFF
		if(MAX_PAGE_SIZE>0&&nextSize>MAX_PAGE_SIZE)
			nextSize=MAX_PAGE_SIZE;
		if(nextSize<size) nextSize=size;
		actualSize=((nextSize+ROUNDOFF+sizeof(MemPageDescriptor)-1)/
			ROUNDOFF)*ROUNDOFF-sizeof(MemPageDescriptor);

		#ifdef MEMORY_H_DEBUG
			reallocNum++;
			memFree=actualSize;
			total+=actualSize;
		#endif

		    // Allocate the new page, test for error, initialize maxPtr
		curPtr=(char *) malloc(actualSize+sizeof(MemPageDescriptor));
		if(!curPtr)
		{
			fprintf(stderr, "CRITICAL ERROR Cannot allocate memory"
				" by the class \'MemAlloc\'\n"
				"\'malloc\' returned NULL after requesting %d "
				"bytes", actualSize+sizeof(MemPageDescriptor));
			exit(1);
		}
		maxPtr=curPtr+actualSize;

		    // Initialize the page descriptor node in the list of pages
		((MemPageDescriptor *)maxPtr)->next=pageList;
		pageList=(MemPageDescriptor *)maxPtr;
		pageList->page=curPtr;

		    // Calculate the size of the next page.
		nextSize=(int)((float)nextSize*GROWTH_RATIO);

		    // Allocate the requested memory.
		temp=curPtr;
		curPtr+=size;

		#ifdef MEMORY_H_DEBUG
			memBusy+=size;
			memFree-=size;
		#endif

		return(temp);
	}


	/*
	void freeAll()
	      Releases (returns to the system) all memory previously allocated
	      by the method 'alloc' of the given object (doesn't affect memory
	      allocated by the other objects of this class).

	      AFTER THE CALL OF THIS FUNCTION, ALL THE POINTERS RETURNED BY THE
	      METHOD 'alloc' BECAME INVALID AND CAN NOT BE DEREFERENCED ANYMORE
	*/
	void freeAll()
	{
		MemPageDescriptor *cur;

		#ifdef MEMORY_H_DEBUG
			restartNum++;
			total=memFree=memBusy=reallocNum=0;
		#endif

			// Release the memory allocated for pages.
		while(pageList)
		{
			cur=pageList;
			pageList=cur->next;
			free(cur->page);
		}

			// Reset the values of the variables
		nextSize=FIRST_PAGE_SIZE;

			// Set maxPtr and curPtr the way that will definitely
			// cause reallocation after the first call of the alloc
		maxPtr=0; curPtr=maxPtr+1;
	}

	#ifdef MEMORY_H_DEBUG
	void printStatistics()
	{
	     printf("Number of \'freeAll\' method calls:   %d\n", restartNum);
	     printf("Number of reallocations:            %d\n", reallocNum);
	     printf("Allocated memory:                   %d bytes\n", memBusy);
	     printf("Remaining size of last buffer:      %d bytes\n", memFree);
	     printf("Total memory allocated for buffers: %d bytes\n ",total);
	}
	#endif

	/*
	MemAlloc(size_t firstPageSize, float growthRatio, size_t maxPageSize)
		CONSTRUCTOR. Defines general allocation policies to be used
		by the object of this class.
		The given values CANNOT be changed later.

	OPTIONAL PARAMETERS:
		firstPageSize -Size of the first page of memory to be allocated.
		                Should be positive number (its type is the same
				as the parameter in the function 'malloc').

		growthRatio  -Size of each new page is the size of the previous
		             page times growthRatio up to the value specified by
		             maxPageSize. The size of actually allocated page
		             may be slightly larger than that calculated by
		             this formula because of ROUNDOFF.
		             growthRatio CANNOT BE LESS THAN 1.

		maxPageSize  -Maximal Size of the page. LESS THAN OR EQUAL TO 0
		              indicates that pages are unlimited. It can't be
		              less than firstPageSize. Page may be larger than
		              the given number because of the following reasons:
		           1. ROUNDOFF
		           2. The program requested a piece of memory which size
					is more than maxPageSize

	*/
	MemAlloc(size_t firstPageSize=MEM_DEF_SIZE, float growthRatio
		=MEM_DEF_FACTOR, size_t maxPageSize=MEM_DEF_MAX):
	    FIRST_PAGE_SIZE(firstPageSize),GROWTH_RATIO(growthRatio),
			MAX_PAGE_SIZE(maxPageSize)
	{
			// Checking for errors (terminate if they occur)
		if(FIRST_PAGE_SIZE<=0)
		{
			fprintf(stderr, "CRITICAL ERROR while constructing the"
				"class \'MemAlloc\':\n"
				"   size of the initial piece of memory "
				"(first parameter of the constructor)\n"
				"   cannot be %d. It should be greater "
				"than 0.\n\n", FIRST_PAGE_SIZE);
			exit(1);
		}
		if(GROWTH_RATIO<1)
		{
			fprintf(stderr, "CRITICAL ERROR while constructing "
				"the class \'MemAlloc\':\n"
				"    growth factor (second parameter of the "
				"constructor) is %.2f\n"
				"    It should be greater than "
				"or equal to 1!\n\n", GROWTH_RATIO);
			exit(1);
		}
		if(MAX_PAGE_SIZE>0&&MAX_PAGE_SIZE<FIRST_PAGE_SIZE)
		{
			fprintf(stderr, "CRITICAL ERROR while constructing"
				" the class \'MemAlloc\':\n"
				"   size of the maximal piece of memory "
				"(third parameter of the constructor)\n"
				"   is %d, which is less than the size of the "
				"initial piece: %d.\n\n",
				MAX_PAGE_SIZE, FIRST_PAGE_SIZE);
			exit(1);
		}
			// Initialize the variables
		nextSize=FIRST_PAGE_SIZE;
		pageList=0;
		maxPtr=0;
			// curPtr is set to be greater than maxPtr
			// so that allocation of the first page will certainly
			// be done during first call of 'alloc'
			// even if its parameter is zero.
			// This is done to avoid returning zero pointer.
		curPtr=maxPtr+1;

		#ifdef MEMORY_H_DEBUG
			total=memFree=memBusy=reallocNum=0;
			restartNum=0;
		#endif

	}

		// Destructor: releases all memory.
	~MemAlloc() {freeAll();}
};

/*
class BaseMemNode

Base class for the classes 'MemNode' and 'MemNodeN'.

IT SHOULD NOT BE USED IN THE USER PROGRAM.
IMPLEMENTED ONLY TO PROVIDE THE COMMON METHODS.
*/
class BaseMemNode
{
protected:
	void **freeList;            // List of free nodes.
	MemPageDescriptor *pageList;// Linked list of internal pages allocated
	                            //  by an object of the class using malloc

	int nextNum;   // Capacity of the page to be allocated next (in nodes)
	const unsigned int NODE_SIZE;// Size of each node CANNOT BE LESS THAN
	                             //  A SIZE OF A POINTER.
	const int START_NODE_NUM; // Capacity of the first page to be allocated
	const float GROWTH_RATIO; // Growth factor (like that in MemAlloc).
	const int MAX_NODE_NUM;   // Maximal capacity of the page

		// Constants for debugging
	#ifdef MEMORY_H_DEBUG
		int nodeFree;
		int nodeBusy;
		int total;
		int reallocNum;
		int restartNum;
	#endif

		// Allocates additional page and adds the nodes from this
		//  page into the free node list.
	void realloc()
	{
		char *curPtr, *maxPtr;
		int nextSize;
		int actualSize;

			// Calculate the size of the page (like in MemAlloc).
		if(MAX_NODE_NUM>0&&nextNum>MAX_NODE_NUM) nextNum=MAX_NODE_NUM;
		nextSize=nextNum*NODE_SIZE;
		actualSize=((nextSize+ROUNDOFF+sizeof(MemPageDescriptor)-1)
			/ROUNDOFF)*ROUNDOFF-sizeof(MemPageDescriptor);

		#ifdef MEMORY_H_DEBUG
			reallocNum++;
			total+=actualSize;
		#endif

			// Allocate the memory, check for errors
		curPtr=(char *) malloc(actualSize+sizeof(MemPageDescriptor));
		if(!curPtr)
		{
			fprintf(stderr, "CRITICAL ERROR Cannot allocate memory "
			    "by the class \'BaseMemNode\' or its subclass:\n"
			    "\'malloc\' returned NULL after requesting %d bytes"
			    , actualSize+sizeof(MemPageDescriptor));
			exit(1);
		}
		maxPtr=curPtr+actualSize;

		    // Initialize the page descriptor node in the list of pages
		((MemPageDescriptor *)maxPtr)->next=pageList;
		pageList=(MemPageDescriptor *)maxPtr;
		pageList->page=curPtr;

			// Calculate the capacity of the next page
		nextNum=(int)((float)nextNum*GROWTH_RATIO);

			// Update the list of free nodes.
		for(maxPtr=curPtr+(actualSize/NODE_SIZE-1)*NODE_SIZE;
			maxPtr>=curPtr;maxPtr-=NODE_SIZE)
		{
			#ifdef MEMORY_H_DEBUG
				nodeFree++;
			#endif

			*(void ***)maxPtr=freeList;
			freeList=(void**)maxPtr;
		}
	}

public:
	/*
		void freeAll()

		Release the memory allocated for the nodes and return it
		  back to the system.
		This releases all memory ever used by the given object
		  of this class.

		'Delete' method, on the other hand, DOES NOT
		return memory to the system
	*/
	void freeAll()
	{
		MemPageDescriptor *cur;

		#ifdef MEMORY_H_DEBUG
			nodeFree=nodeBusy=total=reallocNum=0;
			restartNum++;
		#endif


			// Release the memory used by the pages.
		while(pageList)
		{
			cur=pageList;
			pageList=cur->next;
			free(cur->page);
		}

			// Reset the variables
		freeList=0;
		nextNum=START_NODE_NUM;
	}

	#ifdef MEMORY_H_DEBUG
	void printStatistics()
	{
	     printf("Number of \'freeAll\' method calls:   %d\n", restartNum);
	     printf("Number of reallocations:            %d\n", reallocNum);
	     printf("Size of a node:                     %d\n",NODE_SIZE);
	     printf("Number of allocated nodes:          %d\n", nodeBusy);
	     printf("Number of free nodes:               %d\n", nodeFree);
	     printf("Total memory allocated for buffers: %d bytes\n ",total);
	}
	#endif

	/*
	BaseMemNode(int sizeOfNode, int startNodeNum, float growthRatio,
		int maxNodeNum)

	Constructor. Defines general allocation policies to be used by the
	object of this class. The given values CANNOT be changed later.

	PARAMETERS:
		sizeOfNode - Size of node. Only memory pieces of this size
			will be returned during the memory allocation from
			this class. NOTE: sizeOfNode CANNOT BE LESS THAN
			THE SIZE OF THE POINTER (4 bytes on the most system).

  		startNodeNum - Number of nodes in the first allocated page

		growthRatio  - growth factor like the parameter to the
		               constructor MemAlloc.

		maxNodeNum   - The maximal size of a page (in nodes). If
		               it is 0 or less, page size is unlimited.
		               Cannot be less than startNodeNum.
	*/
	BaseMemNode(unsigned int sizeOfNode, int startNodeNum,
		float growthRatio,int maxNodeNum):   NODE_SIZE(sizeOfNode),
		START_NODE_NUM(startNodeNum), GROWTH_RATIO(growthRatio),
		MAX_NODE_NUM(maxNodeNum)
	{
			// Checking for errors in the parameters.
		if(sizeOfNode<sizeof(void *))
		{
			fprintf(stderr, "CRITICAL ERROR while constructing the"
				" class \'BaseMemNode\' or its subclass:\n"
				"   size of the node cannot be %d. It should be"
				" greater than or equal to the \n"
				"   size of the pointer, i.e. %d.\n\n",
				sizeOfNode, sizeof(void *));
			exit(1);

		}
		if(START_NODE_NUM<=0)
		{
			fprintf(stderr, "CRITICAL ERROR while constructing the "
				"class \'BaseMemNode\' or its subclass::\n"
				"   starting number of nodes cannot be %d.\n"
				"   It should be greater than 0.\n\n",
				START_NODE_NUM);
			exit(1);
		}
		if(MAX_NODE_NUM>0&&MAX_NODE_NUM<START_NODE_NUM)
		{
			fprintf(stderr, "CRITICAL ERROR while constructing the "
				"class \'BaseMemNode\' or its subclass::\n"
				"   the maximal number of nodes is %d. "
				"It CANNOT be less than\n"
				"   the starting number of nodes which is %d."
				"\n\n", MAX_NODE_NUM,START_NODE_NUM);
			exit(1);
		}
		if(GROWTH_RATIO<1)
		{
			fprintf(stderr, "CRITICAL ERROR while constructing the "
				"class \'BaseMemNode\' or its subclass:\n"
				"    growth factor is %.2f\n"
				"    It should be greater than or equal to 1."
				"\n\n", GROWTH_RATIO);
			exit(1);
		}
			// Initialize variables
		nextNum=START_NODE_NUM;
		pageList=0;
		freeList=0;

		#ifdef MEMORY_H_DEBUG
			nodeFree=nodeBusy=total=reallocNum=restartNum=0;
		#endif
	}

		// Destructor: releases all memory.
	~BaseMemNode(){freeAll();}
};

/*
class MemNodeN

Provides fast methods for allocating/releasing memory of the fixed sizes,
such as nodes of linked lists, trees, etc. The size of a node is given as
a parameter to the constructor and cannot be changed afterwards (to use
different nodes, use several objects of this class).

The 'New()' method is used to allocate a node (returns void *).
The 'Delete(void *)' method is used to release the node, i.e. add the node to
the list of free nodes so it can be allocated by the subsequent call of 'New()'

WARNING: THIS METHOD IS NOT LESS VULNERABLE THAN FUNCTION 'free' OR OPERATOR
'delete'. GIVING AS A PARAMETER A VALUE WHICH WAS NOT RETURNED BY THE
PREVIOUS 'New' CALL IS A CRITICAL ERROR WHICH CANNOT BE DETECTED.
IT CAN LEAD TO UNPREDICTABLE RESULTS.

NOTE: The method 'Delete' doesn't return the memory to the system.
The memory cannot be reused in any other part of the program including
other objects of this class.

The method 'freeAll' DOES return ALL the allocated memory to the system.
*/
class MemNodeN: public BaseMemNode
{
public:
	/*
	MemNodeN(int sizeOfNode, int startNodeNum, float growthRatio,
		int maxNodeNum)

	Constructor. The same as the constructor of BaseMemNode, except
		some parameters are optional.

	REQUIRED PARAMETER: 	sizeOfNode
	OPTIONAL PARAMETERS:    startNodeNum, growthRatio, maxNodeNum
	*/
	MemNodeN(int sizeOfNode, int startNodeNum=NODE_DEF_NUM,
		float growthRatio=MEM_DEF_FACTOR, int maxNodeNum=NODE_DEF_MAX):
		BaseMemNode(sizeOfNode, startNodeNum, growthRatio, maxNodeNum)
	{}

	/*
	void *New()

	  Allocates a new node and returns a pointer to it. Node's size is
	  given as a parameter of the constructor and cannot be changed.
	*/
	void *New()
	{
		#ifdef MEMORY_H_DEBUG
			nodeBusy++;
			nodeFree--;
		#endif

		void *temp;
		if(!freeList) realloc();


		temp=(void *)freeList;
		freeList=(void **)(*freeList);
		return(temp);
	}

	/*
	void Delete(void *Node)

	  Adds a given node to the list of free nodes. In contrast with
	  'freeAll' method the memory is not given back to the system.

	PARAMETERS:
		Node - Pointer to the node to deallocate. SHOULD BE A VALUE
		       RETURNED BY THE METHOD 'New', PROVIDED THERE WERE NO
		       CALLS TO THE METHOD 'freeAll' AFTER THAT CALL OF 'New'.
	*/
	void Delete(void *Node)
	{
		#ifdef MEMORY_H_DEBUG
			nodeBusy--;
			nodeFree++;
		#endif

		*(void ***)Node=freeList;
		freeList=(void **)Node;
	}
};


/*
template <class Node> class MemNode

PARAMETERS:
	Node	-  A structure which represents the node for the class
			to allocate and deallocate.

DESCRIPTION:
  Performs the same functions as MemNodeN, but has an advantage of possibility
  to be constructed for a specific node. This eliminates the necessity to
  calculate the size of the node (to give it as a parameter to the constructor),
  and convert the pointer after a call to the 'New' function.

  All comments of MemNodeN are applicable to this class, except the 'New' and
  'Delete' work with (Node *) pointers instead of (void *).

  NODE: The structure given as a parameter should have the global scope.
*/
template <class Node> class MemNode: public BaseMemNode
{
public:
	/*
	MemNode(int startNodeNum, float growthRatio, int maxNodeNum)

	Constructor. The same as the constructor of BaseMemNode, except
		first parameter is missing and the rest is OPTIONAL.
	*/
	MemNode(int startNodeNum=NODE_DEF_NUM, float growthRatio=MEM_DEF_FACTOR,
	   int maxNodeNum=NODE_DEF_MAX):
	   BaseMemNode(
	       (sizeof(Node)>=sizeof(void *))?sizeof(Node):sizeof(void *),
	       startNodeNum, growthRatio, maxNodeNum)
	{}

	  // Works the same as MemNode::New
	Node *New()
	{
		#ifdef MEMORY_H_DEBUG
			nodeBusy++;
			nodeFree--;
		#endif

		Node *temp;
		if(!freeList) realloc();

		
		temp=(Node *)freeList;
		freeList=(void **)(*freeList);
		return(temp);
	}

		// Works the same as MemNode::Delete
	void Delete(Node *node)
	{
		#ifdef MEMORY_H_DEBUG
			nodeBusy--;
			nodeFree++;
		#endif

		*(void ***)node=freeList;
		freeList=(void **)node;
	}
};

#endif
