/*==========================================================================;
 *
 *  AUTHOR:    Iakhiaev, Mikhail.
 *
 *  FILE NAME: rulesact.cpp
 *
 *  CONTENTS:  Part of the FRONT END, PSGRND.
 *             Partial implementation of the class 'Rules'
 *             Contains all the aciton handlers (i.e. methods called from
 *              the file "grammar.y") and only them. Neither of these
 *              actions prints error messages.
 *
 *  DATE:     March 24, 2004
 ***************************************************************************/

#include "rules.h"

/*
void Rules::ruleNew(TempAtom *antecedent, TempAtom *consequent, int horn)
DESCRIPTION
	Reduce lists of atoms to a rule. This method checks the rule for
	errors, creates local and global quantifiers, and converts the lists
	of atoms from temporary structures to permanent

PARAMETERS
	antecedent    -  List of atoms in the body, can be NULL
	consequent    -  List of atoms in the head, can be NULL
	horn          -  TRUE if we are reducing a horn rule,
	                 FALSE if we are reducing implication
*/
void Rules::ruleNew(TempAtom *antecedent, TempAtom *consequent, int horn)
{
	TempAtom *atom;
	TempAtom *condList=0;        // List of conditions to be created
	Quantifier *globalQuant=0;   // List of global quantifiers to be created
	RuleNode *rule, *curRule;
	RuleNode dummyRule;
			// Get the line number of the rule
	int lineNum = lexan.getLineNum(isParserLookahead());

		// If debugging is enabled, print the rule before processing:
	if(DEBUG_PRINT_RULE_BEFORE_PROCESSING)
	{
		printf("BEFORE(%d): ", lineNum);
		printTempRule(antecedent, consequent, horn);
	}

		// Check and report for data atoms:
		//    "dataType(variableOfThisDataType)"
	reportUselessData(antecedent);
	reportUselessData(consequent);

		// If this is a horn rule, check for horn predicates
		//  in cardinality lists:
	if(horn && !checkHornRule(antecedent, consequent))
	{
		deleteTempAtomList(antecedent);
		deleteTempAtomList(consequent);
		return;
	}

		// Remove atoms: ATOM_TRUE, ATOM_FALSE.
		//  In some cases, the rule might be eliminated.
	if(!removeTrueFalseAtoms(&antecedent, 0)||
		!removeTrueFalseAtoms(&consequent, 1))
	{
		deleteTempAtomList(antecedent);
		deleteTempAtomList(consequent);
		return;
	}
		// Check consistency in the local variable lists:
	checkLocalVarLists(antecedent, consequent);
		// Determine which variables are local and which are global
	determineLocalGlobal(antecedent, consequent);
		// Replace underscores with temporary variables:
	processUnderscores(antecedent, consequent);

		// Construct a list of global variables.
		// Separate data and predefined atoms from the rest and put
		//   them into condition list.
		// Create data atoms out of predicate tails
		// Convert arithmetic expressions in data or predefined atoms
		//  to assignments
	globalQuant=createConditions(&antecedent, &consequent, &condList);

		// Convert existential atoms in the body to cardinality atoms:
	convertBodyExistentialAtoms(&antecedent);

		// If debugging is enabled, print the rule in the
		//  intermediate stage:
	if(DEBUG_PRINT_RULE_INTERMEDIATE)
	{
		printf("INTERMEDIATE(%d): [", lineNum);
		printQuantifierList(globalQuant);
		printf(" ( ");
		for(atom=condList;atom;atom=atom->next)
			printTempAtom(atom);
		printf(" )] ");

		printTempRule(antecedent, consequent, horn);
	}

		// Build local quantifiers from the list of local variables and
		//  the list of conditions from data restrictions:
	for(goAtoms.start(antecedent, consequent);(atom=goAtoms.next());)
	{
		atom->prog.quant=quantBuilder.build(atom->prog.restrictions,
		                                    atom->prog.quant);
		atom->prog.restrictions=0;
	}

		// Build global quantifiers from the list of global variables
		//  and the list of conditions (obtained by 'createConditions')
	globalQuant=quantBuilder.build(condList, globalQuant); condList=0;

	// after global quantifiers have been build
	if(DEBUG_PRINT_RULE_INTERMEDIATE)
	{
		printf("INTER(%d): [", lineNum);
		printQuantifierList(globalQuant);
		printf(" ( ");
		printf(" )] ");

		printTempRule(antecedent, consequent, horn);
	}
		// Allocate and initialize the rule node
	rule=ruleSrc.New();
	rule->maxGroundRule=0;
	rule->next=0;
	rule->horn=horn;
	rule->lineNum=lineNum;
	rule->quant=globalQuant;

		// Convert temporary atoms to permanent and finish
		//  initializing the rule node.
	rule->maxAtomNum =
	        convertTempAtomList(antecedent, &(rule->body),
	                          &(rule->cardBody), &(rule->weightBody))
	                 +
	        convertTempAtomList(consequent, &(rule->head),
	                          &(rule->cardHead), &(rule->weightHead));

		// Append the rule to the right place in the rule sequence
		//  RIGHT PLACE: The next rule will be the first that has
		//   estimated maximum number of atoms greater than the
		//  estimated maximum number for the current rule

	dummyRule.next=ruleSequence;
		// Find the right place in the rule sequence:
	for(curRule=&dummyRule ;
	    curRule->next &&  curRule->next->maxAtomNum <= rule->maxAtomNum ;
	    curRule = curRule->next);

		// Append the current rule to that place:
	rule->next=curRule->next;
	curRule->next=rule;
	ruleSequence=dummyRule.next;
}


// This macro is a common step in creating any temporary atoms. It
//  allocates the node and initializes the references.
#define NEW_TMP_ATOM(a) a=(TempAtom *)tempSrc.New();a->next=0;a->last=a;


/*
TempAtom *Rules::metaAtomNew(TempAtom *cList, ExpressionHandler *ehMin,
                             ExpressionHandler *ehMax, int atomType)
DESCRIPTION
	Reduce meta-list and meta-limits to a meta-atom. The method does
	similar things for both cardinality and weight atoms

PARAMETERS
	cList         -  Meta-list: List of program atoms
	ehMin, ehMax  -  Expression handlers for the meta-limits
	atomType      -  Either ATOM_CARD or ATOM_WEIGHT
RETURNS
	Newly allocated cardinality or weight atom, ATOM_TRUE, or ATOM_FALSE
*/
#define RETURN_ATOM(atom, trueFalse) \
	{deleteTempAtomList(cList); \
	deleteExpression((atom)->meta.min); \
	deleteExpression((atom)->meta.max); \
	(atom)->type=(trueFalse); \
	return(atom);}
TempAtom *Rules::metaAtomNew(TempAtom *cList, ExpressionHandler *ehMin,
                             ExpressionHandler *ehMax, int atomType)
{
	TempAtom *atom;
	Expression *exp;
	char checkMin=0, checkMax=0;

		// Create temporary atom node and initialize the references:
	NEW_TMP_ATOM(atom);
	atom->meta.min=0;
	atom->meta.max=0;

	// CHECK POSSIBLE CASES FOR THE LOWER AND UPPER BOUND:

	if(ehMin) // If the lower bound was given:
		if(IS_EXP_VARNUM(ehMin->type))
		{
			// If the expression has the correct type:
				// Attach the expression to the atom:
			atom->meta.min=ehMin->first;
				// If number, can do more checks (later)
			if(ehMin->type==ET_NUM) checkMin=1;
			tempSrc.Delete(ehMin);  // Release the handler
		}
		else
		{
				// Wrong type of expression -> report an error
				//  and release the expression:
			if(IS_EXP_SAY_ERROR(ehMin->type))
			    semanticError(SE_META_BAD_LIMIT, atomType,0 ,ehMin);
			deleteExpression(ehMin);
		}
	else // If there is no lower bound:
		if(!ehMax)
		{
			// If there is neither the lower nor the upper bound,
			//  report a warning:
			semanticError(SW_META_NO_LIMITS,atomType,cList);
			RETURN_ATOM(atom, ATOM_TRUE)
		}


	if(ehMax)  // If the upper bound was given:
	{
		if(IS_EXP_VARNUM(ehMax->type))
		{
			// If the type is correct:
				// Attach the expression to the atom
			atom->meta.max=ehMax->first;
				// If number, can do more checks (later)
			if(ehMax->type==ET_NUM) checkMax=1;
			tempSrc.Delete(ehMax); // Release the handler
		}
		else
		{
			// Wrong type of expression -> report an error and
			//  release the expression:
			if(IS_EXP_SAY_ERROR(ehMax->type))
			   semanticError(SE_META_BAD_LIMIT, atomType, 1, ehMax);
			deleteExpression(ehMax);
			if(!atom->meta.min) RETURN_ATOM(atom, ATOM_TRUE)
		}
	}

		// If the lower bound was not given, it should be equal to zero
		// Allocate an expression node and initialize it correspondingly
	if(!atom->meta.min)
	{
		exp=expSrc.New();
		exp->next=0;
		exp->token=EXPR_NUM;
		exp->num=0;
		atom->meta.min=exp;
		checkMin=1;          // Can check the lower bound later
	}

		// If the upper bound was not given, it should be equal to
		// the maximum possible integer.
		// Allocate an expression node and initialize it correspondingly
	if(!atom->meta.max)
	{
		exp=expSrc.New();
		exp->next=0;
		exp->token=EXPR_NUM;
		exp->num=INTEGER_MAX;
		atom->meta.max=exp;
		checkMax=1;           // Can check the upper bound later
	}

	// DO ADDITIONAL CHECKS FOR WARNINGS IF WE CAN:

		// Upper limit less than zero => atom is always false
	if(checkMax&&atom->meta.max->num<0)
	{
		semanticError(SW_META_NEGATIVE,atomType,1,atom->meta.max->num);
		RETURN_ATOM(atom, ATOM_FALSE)
	}

		// Lower limit less than zero doesn't make any sense:
	if(checkMin&&atom->meta.min->num<0)
	{
		semanticError(SW_META_NEGATIVE,atomType,0,atom->meta.min->num);
		atom->meta.min->num=0;
	}

	if(checkMin&&checkMax)
	{
			// Upper limit is less than lower limit =>
			//   atom is always false
		if(atom->meta.min->num > atom->meta.max->num)
		{
			semanticError(SW_META_MAX_LESS_MIN, atomType,
			             atom->meta.min->num, atom->meta.max->num);
			RETURN_ATOM(atom, ATOM_FALSE)
		}

			// Lower limit is zero and the upper limit is the
			//  maximum possible integer => atom is always true
		if(atom->meta.min->num == 0 && atom->meta.max->num==INTEGER_MAX)
		{
			semanticError(SW_META_TRUE, atomType);
			RETURN_ATOM(atom, ATOM_TRUE)
		}
	}

		// Indicate atom type and attach meta-list to the atom:
	atom->type=atomType;
	atom->meta.list=cList;

	return(atom);
}


/*
TempAtom *Rules::programAtomFromSimple(
         TempAtom *simpleAtom, Quantifier *localVars, TempAtom *restrictions)
DESCRIPTION
	Converts simple atom (program atom without local variables and data
	restrictions) to the general case of program atom. Even though
	some general program atoms can be constructed without this stage,
	the method is necessary due to specifics of the grammmar.

PARAMETERS
	simpleAtom     -  Simple program atom that needs to be converted
	localVariables -  List of local variables to be added to the atom,
	                   can be NULL
	restrictions   -  Data restrictions to be added to the atom, can be NULL
RETURNS
	The same atom as given in 'simpleAtom'
*/
TempAtom *Rules::programAtomFromSimple(
         TempAtom *simpleAtom, Quantifier *localVars, TempAtom *restrictions)
{
	if(simpleAtom->type==ATOM_PROG)
	{
		// If atom is in fact the program atom, attach the lists
		simpleAtom->prog.restrictions=restrictions;
		simpleAtom->prog.quant=localVars;
	}
	else
	{
		// If atom is not the program atom (possible case: ATOM_FALSE),
		//  i.e. error occured on the previous stage,
		//  deallocate the lists
		deleteTempAtomList(restrictions);
		deleteQuantifierList(localVars);
	}

	return(simpleAtom); // Return the same atom as was given
}


/*
TempAtom *Rules::weightAssignmentFromSimple(
                       TempAtom *simpleAtom, ExpressionHandler *eh,
                       Quantifier *localVars, TempAtom *restrictions)
DESCRIPTION
	Creates weight assignment by adding local variables, data restrictions
	and weight expression to a simple atom

PARAMETERS
	SimpleAtom    -  Atom to be converted
	eh            -  Weight expression
	localVars     -  Local variables, can be NULL
	restrictions  -  Data restrictions, can be NULL
RETURNS
	The same atom as given in the 'simpleAtom' with attached lists,
	unless there was an error in which case it is ATOM_FALSE

*/
TempAtom *Rules::weightAssignmentFromSimple(
                       TempAtom *simpleAtom, ExpressionHandler *eh,
                       Quantifier *localVars, TempAtom *restrictions)
{
		// If the atom is not a program atom (possible case: ATOM_FALSE)
		//  Deallocate all the lists and return the atom
	if(simpleAtom->type!=ATOM_PROG)
	{
		deleteExpression(eh);
		deleteTempAtomList(restrictions);
		deleteQuantifierList(localVars);
		return(simpleAtom);
	}

		// Attach the restrictions and local variables to the atom:
	simpleAtom->prog.restrictions=restrictions;
	simpleAtom->prog.quant=localVars;

	if(!eh) // No expression -> critical error
	{
		fprintf(stderr, "CRITICAL ERROR: during creating "
			"weight assignment. The expression\n"
			"at the right side of \'=\' is not given.\n");
		exit(1);
	}

		// If the type of expression is correct, attach it to
		//  the atom and return the atom
	if(IS_EXP_WEIGHT(eh->type))
	{
		simpleAtom->prog.weight=eh->first;
		tempSrc.Delete(eh);
		return(simpleAtom);
	}

	// WE CAN GET HERE ONLY IN CASE OF ERROR:

		// Report the error if necessary:
	if(IS_EXP_SAY_ERROR(eh->type))
		semanticError(SE_WEIGHTEXP_BAD, simpleAtom, eh);

		// Release all unnecessary memory.
	deleteExpression(eh);
	deleteTempAtomList(restrictions);
	deleteQuantifierList(localVars);
	deleteExpression(simpleAtom->prog.arg);

		// Return the FALSE atom:
	simpleAtom->type=ATOM_FALSE;
	return(simpleAtom);
}


/*
ExpressionHandler *Rules::weightCallNew(int symbolno,ExpressionHandler *argList)
DESCRIPTION
	Create a weight function call.

PARAMETERS
	symbolno  -  Symbol number of the weight function to be called
	argList   -  List of arguments of the weight function
RETURNS
	Expression handler containing the expression
	with the weight function call
*/
ExpressionHandler *Rules::weightCallNew(int symbolno,ExpressionHandler *argList)
{
	ExpressionHandler *eh, *temp;
	WFnode *wn;
	Expression *exp;
	Expression *first, *last;

		// Allocate the new expression handler node and initialize it
		//  by default as error expression:
	eh=(ExpressionHandler *)tempSrc.New();
	eh->next=0;
	eh->first=eh->last=0;
	eh->type=ET_ERROR;

		// Cannot make weight function call unless finished
		//  parsing data sequence.
	if(!isDataFinished)
	{
			// Report semantic error
		semanticError(SE_WEIGHT_IN_DECLARATION, symbolno);

			// Release the argument list:
		for(;argList;temp=argList,argList=argList->next,
			deleteExpression(temp));

		return(eh); // Return the error expression
	}

		// Check for critical errors:
	if(symbt.getType(symbolno)!=TYPE_WEIGHT_FUN)
	{
		fprintf(stderr, "CRITICAL ERROR: during "
		 "creating weight function call\n"
		 "Expected weight function but the type of identifier was %d\n",
		    symbt.getType(symbolno));
		exit(1);
	}

	wn=(WFnode *)((symbt.getVal(symbolno))->ptr);

		// Create arguments for the weight function. If this is failed,
		//  return the error expression.
	if(!(first=createArguments(argList, symbolno))) return(eh);

		// Find the last argument in the argument list:
	for(last=first;last->next;last=last->next);

		// Initialize the new expression handler properly:
	eh->type=ET_WEIGHT;
	eh->first=first;
		// Allocate the new expression node and append it to the
		//  end of the expression list:
	eh->last = last->next = exp = expSrc.New();

		// Initialize the new node with the weight function call:
	exp->next=0;
	exp->token=EXPR_WEIGHT;
	exp->wFun=wn;

	return(eh);
}


/*
TempAtom *Rules::predAtomNew(ExpressionHandler *HLeft,
                             ExpressionHandler *HRight, int relOp)
DESCRIPTION
	Given two operands and relational operand code, this method creates
	a predefined atom.

PARAMETERS
	HLeft, HRight - left and right operands
	relOp         - relational operator code

RETURNS
	Newly constructed predefined atom or ATOM_FALSE
*/
TempAtom *Rules::predAtomNew(ExpressionHandler *HLeft,
                             ExpressionHandler *HRight, int relOp)
{
	TempAtom *atom;
	Expression *newExp;
	int combinedType;
	int cmpFlag;

	NEW_TMP_ATOM(atom) // Allocate the new atom node

	combinedType=HRight->type|HLeft->type;

		// Check whether the types of operands are appropriate to
		//  construct a predefined atom. If one of the operands
		//  is a weight function, we cannot do this.
	if((combinedType&ET_WEIGHT) || (combinedType&ET_ERROR))
	{
			// Report an error if it has not been reported yet:
		if(IS_EXP_SAY_ERROR(combinedType))
			semanticError(SE_PATOM_WRONGTYPES, HLeft, HRight);

			// Release the expression nodes and return FALSE atom.
		deleteExpression(HLeft);
		deleteExpression(HRight);
		atom->type=ATOM_FALSE;
		return(atom);
	}

		// If one of the operands can be a string and the other can be a
		//  number, we are handling a special case
	if((EXP_CAN_BE_STRING(HLeft->type) && EXP_CAN_BE_NUMBER(HRight->type))||
	    (EXP_CAN_BE_NUMBER(HLeft->type) && EXP_CAN_BE_STRING(HRight->type)))
	{
		// Check whether one of the operands has to remain number in any
		//  case, even when the number is in the range for constants:
		if(!EXP_CAN_BE_STRING(HLeft->type))   cmpFlag=EXPR_CMP_LEFTNUM;
		else if(!EXP_CAN_BE_STRING(HRight->type))
		                                      cmpFlag=EXPR_CMP_RIGHTNUM;
		     else cmpFlag=EXPR_CMP_SPECIAL; // This is just special case
	}
	else	// Otherwise, nothing special in this comparison.
		cmpFlag=0;

		// CONSTRUCT THE PREDEFINED ATOM:
		// Allocate and initialize the new expression node:
	newExp=expSrc.New();
	newExp->next=0;
	newExp->token=relOp;
	newExp->num=cmpFlag;	// cmpFlg is saved in the num field


		// Merge the operands and append the allocated node to the end:
	HLeft->last->next=HRight->first;
	HRight->last->next=newExp;

		// Append the resulting expression to the atom node:
	atom->type=ATOM_CONDITION;
	atom->cond.first=HLeft->first;
	atom->cond.last=newExp;

		// Release the expression handlers:
	tempSrc.Delete(HLeft);
	tempSrc.Delete(HRight);

	return(atom);
}


/*
TempAtom *Rules::dataAtomNew(int symbolno, ExpressionHandler *argList)
DESCRIPTION
	Create a new data atom

PARAMETERS
	symbolno  -  Symbol number of the data predicate in the atom
	argList   -  List of arguments of the data predicate
RETURNS
	The newly created data atom or ATOM_FALSE
*/
TempAtom *Rules::dataAtomNew(int symbolno, ExpressionHandler *argList)
{
	TempAtom *atom;
	DPnode *dp=(DPnode *)((symbt.getVal(symbolno))->ptr);

		// Create temporary atom node and initialize references:
	NEW_TMP_ATOM(atom);

		// Check whether the symbolno corresponds to data predicate
	if(symbt.getType(symbolno)!=TYPE_DATA_PRED)
	{
		fprintf(stderr, "CRITICAL ERROR: during creating data atom\n"
			"Expected data predicate, but the type was %d\n",
			symbt.getType(symbolno));
		exit(1);
	}

		// Create arguments. If error, return ATOM_FALSE
	if(!(atom->data.arg=createArguments(argList, symbolno)))
	{
		atom->type=ATOM_FALSE;
		return(atom);
	}

		// Initialize the atom node:
	atom->type=ATOM_DATA;
	atom->data.dp=dp;

	return(atom);
}


/*
TempAtom *Rules::programAtomNew(int symbolno, ExpressionHandler *argList,
                                Quantifier *localVars, TempAtom *restrictions)
DESCRIPTION
	Create a new program atom

PARAMETERS
	symbolno     -  Symbol number of the program predicate for the atom
	argList      -  List of arguments, can be NULL
	localVars    -  List of local variables, can be NULL
	restrictions -  List of data restrictions associated with the atom,
	                  can be NULL
RETURNS
	Newly created program atom or ATOM_FALSE
*/
TempAtom *Rules::programAtomNew(int symbolno, ExpressionHandler *argList,
                                Quantifier *localVars, TempAtom *restrictions)
{
	TempAtom *atom;
	PPnode *pp=(PPnode *)((symbt.getVal(symbolno))->ptr);

		// Create temporary atom node and initialize the references:
	NEW_TMP_ATOM(atom);

		// Check whether the symbolno corresponds to program predicate
	if(symbt.getType(symbolno)!=TYPE_PROGRAM_PRED)
	{
		fprintf(stderr, "CRITICAL ERROR: during creating "
			"simple or existential atom\n"
			"Expected program predicate, but the type was %d\n",
			symbt.getType(symbolno));
		exit(1);
	}

		// Mark this program predicate as used:
	pp->usage=PP_USED;

		// Create arguments. If the arity of the program predicate is
		//  greater than zero, but 'createArguments' returned NULL,
		//  there was an error while creating arguments =>
		//  deallocate the lists and return ATOM_FALSE
	if(!(atom->prog.arg=createArguments(argList, symbolno)) && pp->arity>0)
	{
		deleteTempAtomList(restrictions);
		deleteQuantifierList(localVars);
		atom->type=ATOM_FALSE;
		return(atom);
	}

		// Initialize the program pedicate node:
	atom->prog.pp=pp;
	atom->prog.restrictions=restrictions;
	atom->prog.quant=localVars;
	atom->prog.weight=0;
	atom->type=ATOM_PROG;

	return(atom);
}


/*
Quantifier *Rules::localVarAdd(Quantifier *varList, int newVar)
DESCRIPTION
	Add local variable to the local variable list. This is used to
	parse local variable lists.

PARAMETERS
	varList  -  List of local variables, can be NULL for the first variable
	newVar   -  Symbol number of the new variable to add to the list
RETURNS
	List of program variables with added variable
*/
Quantifier *Rules::localVarAdd(Quantifier *varList, int newVar)
{
	Quantifier *quant;

	    // Check whether number given in the parameter is indeed a symbol
	    //  number of some program variable. Critical error otherwise.
	if(symbt.getType(newVar)!=TYPE_VAR)
	{
		fprintf(stderr, "CRITICAL ERROR: Expected program variable in a"
		    " local variable list\n received identifier of type %d\n\n",
		      symbt.getType(newVar));
		exit(1);
	}

	    // Create and initialize the new quantifier node:
	quant=quantSrc.New();
	quant->prev=quant->next=0;
	quant->token=QUANT_VAR;
	quant->var=(VarNode *)(symbt.getVal(newVar)->ptr);
	quant->conditions=0;

	    // Indicate that the variable has been used:
	quant->var->valIndex=VAR_FREE;

	    // Append this new node to the list given as a parameter
	quant->next=varList;

	    // Return the new head of the list.
	return(quant);
}


/*
ExpressionHandler *Rules::argUndMerge(ExpressionHandler *ALeft,
                                      ExpressionHandler *ARight)
DESCRIPTION
	Merges the two argument lists and adds an underscore between them.
	This method is called for the first underscore that occurrs in
	some argument list (the subsequent underscores are added in the
	same way as the other parameters). The method is necessary
	for 'argumentList' to become 'existentialList'.
PARAMETERS
	ALeft   -  Argument list before the first underscore, can be NULL
	ARight  -  Argument list after the first underscore, can be NULL
NOTE
	The argument lists are given in the REVERSE ORDER. The result
	should also be in the reverse order. 'ARight' should be appended
	the first and 'ARight' the second
RETURNS
	Resulting arugment list
*/
ExpressionHandler *Rules::argUndMerge(ExpressionHandler *ALeft,
                                      ExpressionHandler *ARight)
{
	ExpressionHandler *cur;
		// Allocate the expression handler for
		//  the underscore in the middle
	ExpressionHandler *AMiddle=(ExpressionHandler *)tempSrc.New();
		// Allocate the expression for the underscore
	Expression *expMiddle=expSrc.New();

		// Initialize the handler
	expMiddle->next=0;
	expMiddle->token=EXPR_UNDERSCORE;

		// Initialize the expression
	AMiddle->type=ET_VAR;
	AMiddle->first=AMiddle->last=expMiddle;
		// Append the 'ALeft' to the handler
	AMiddle->next=ALeft;

	if(ARight)
	{ // If 'ARight was given:
			// Find the last element of the 'ARight'
		for(cur=ARight;cur->next;cur=cur->next);
			// Append the handler with the underscore to the
			//  last element of the 'ARight'
		cur->next=AMiddle;
	}
	else
		ARight=AMiddle;  // Do this to return AMiddle

	return(ARight);
}


/*
ExpressionHandler *Rules::argAdd(ExpressionHandler *argList,
                                 ExpressionHandler *newExp)
DESCRIPTION
	Add the new argument to the argument list
NOTE
	Argument list is constructed in the reverse order

PARAMETERS
	argList  -  Argument list, can be NULL
	newExp   -  The new argument to append, can be NULL which means
	             the argument is UNDERSCORE
RETURNS
	The updated argument list
*/
ExpressionHandler *Rules::argAdd(ExpressionHandler *argList,
                                 ExpressionHandler *newExp)
{
	Expression *exp;

		// If 'newExp' is NULL, create the expression corresponding
		//  to underscore
	if(!newExp)
	{
			// Allocate and initialize the new expression node:
		exp=expSrc.New();
		exp->next=0;
		exp->token=EXPR_UNDERSCORE;

			// Allocate and initialize the new handler node:
		newExp=(ExpressionHandler *)tempSrc.New();
		newExp->type=ET_VAR;
		newExp->first=newExp->last=exp;
	}

		// Add the expression to the list and return the new list:
	newExp->next=argList;

	return(newExp);
}


/*
void Rules::rebuild()
DESCRIPTION
	This action is called after 'dataSequence' has been parsed
	but before parsing the declaration sequence.
	It rebuilds the structures for string constants, data predicates,
	and weight functions. It assigns numbers to the string constants
	and converts lists into sorted arrays.
*/
void Rules::rebuild()
{
	WFnode *wNode;
	Value *value, *valList;
	int size, i;
	int isWeight;
	int minNum, constNum;

		// Instances of Array classes to do sorting:
	IntegerArray integerArray;
	DWVectorArray dwVectorArray;

		// Indicate that we finished data sequence:
	isDataFinished=1;

		//  ASSIGNING NUMBERS TO THE CONSTANTS:
	i=strConst.buildReferenceStructure(&minNum, &constNum);
	if(i!=SE_NO_ERROR)
	{
		semanticError(i, minNum, constNum);
		return;	// In the next step, the parser is supposed to stop.
	}

	// REBUILD DATA PREDICATES AND WEIGHT FUNCTIONS:
		// Go over every symbol in the symbol table:
	for(wNode=first.data ; wNode ; wNode=wNode->next)
	{
		isWeight=(symbt.getType(wNode->symbolno) == TYPE_WEIGHT_FUN);

			// Save the value list:
		valList=value=wNode->tmpArguments;

			// Allocate the memory to store the arguments:
		size=wNode->arity*wNode->valNum;
		wNode->arguments=(int *)permMSrc.alloc(size*sizeof(int));

		    // Copy the arguments from the list of values to the array:
		    //  If some argument is a constant, we need to replace the
		    //  symbol number stored in the 'value' node with the number
		    //  assigned to that constant:
		for(i=0; value && i<size; i++, value=value->next)
			if(value->type==EXPR_NUM)
				wNode->arguments[i] = value->num;
			else
				if(value->type==EXPR_STRING)
				    wNode->arguments[i] = value->num =
				      strConst.getNumberForConstant(value->num);
				else
				{
					fprintf(stderr, "CRITICAL ERROR in "
						"Rules::rebuild, detected wrong"
						" type of value\n\n");
					exit(1);
				}

			// Check for critical errors:
		if(value || i!=size)
		{
			fprintf(stderr, "CRITICAL ERROR in Rules::rebuild: "
				"there was an error while "
				"parsing the dataSequence\n\n");
			exit(1);
		}

			// If we are dealing with weight function,
			//  copy the results into the separate array
		if(isWeight)
		{
			value=wNode->tmpResults; // Save the pointer to the list
				// Allocate array for the results
			wNode->results=(int *)permMSrc.alloc(
					wNode->valNum*sizeof(int));

				// Copy the list to the array:
			for(i=0; value && i<wNode->valNum;i++,value=value->next)
				wNode->results[i]=value->num;

			if(value || i!=wNode->valNum)
			{
				fprintf(stderr, "CRITICAL ERROR in "
					"Rules::rebuild: there was an error "
					"while parsing the dataSequence\n\n");
				exit(1);
			}
		}

			// Sort resulting arrays:
		if(!isWeight && wNode->arity==1)
		{
			integerArray.insert(wNode->arguments, wNode->valNum);
			quickSort<IntegerArray>(integerArray);
		}
		else
		{
			dwVectorArray.insert(wNode->arguments, wNode->arity,
				wNode->valNum, (isWeight)?wNode->results:0);
			quickSort<DWVectorArray>(dwVectorArray);
		}

			// Checking for repetitions:
		checkRepetitionsInData(wNode, valList);
	}

		// Release all the value nodes, unless debugging
	if(!DEBUG_DONTFREE)
		valSrc.freeAll();
}


/*
VarNode *Rules::varAdd(VarNode *list, int symbolno)
DESCRIPTION
	Add a variable to the list corresponding to the single
	declaration.

PARAMETERS
	list   -  single linked list of the variables in the opposite order
	          comparing with how they occur in declarations
	          This parameter can be NULL, if the variable is the
	          first one for a given declaration
	symbolno -  symbol number of the variable identifier.

RETURNS
	Pointer to the modified list of variables corresponding to a declaration
*/
VarNode *Rules::varAdd(VarNode *list, int symbolno)
{
	VarNode *vnode;

		// Allocate the node, indicate the identifier as a variable and
		//  link the node to the symbol table:
	vnode=varSrc.New();
	symbt.setType(symbolno,	TYPE_VAR);
	(symbt.getVal(symbolno))->ptr=vnode;

		// Initialize the node:
	vnode->symbolno=symbolno;
	vnode->tmpQuant=0;
	vnode->valIndex=VAR_NOTUSED;
	vnode->type=&DPError;

		// Add the node to the list:
	vnode->next=list;
	return(vnode);
}


/*
void Rules::varDeclarationNew(int data, VarNode *vars)
DESCRIPTION
	This function finishes variable declaration. For a list constructed by
	multiple calls of 'varAdd', this method sets appropriate data type to
	every variable and adds these variables to the global
	list of program variables

PARAMETERS
	data   -  symbol number of the data type of the variables
	vars   -  single linked list of the variables in the opposite order
	          comparing with how they occur in the declaration
*/
void Rules::varDeclarationNew(int data, VarNode *vars)
{
	VarNode *cur, *src, *dest;
	DPnode *dp=(DPnode *)((symbt.getVal(data))->ptr);

		// Check for critical errors:
	if(symbt.getType(data)!=TYPE_DATA_PRED)
	{
		fprintf(stderr, "CRITICAL ERROR: while parsing "
			"variable declaration \n"
			"Expected data predicate, but the type was %d\n",
				symbt.getType(data));
		exit(1);
	}

		// Check whether the data predicate can possibly be
		// a type of the variables (only unary data predicates can)
	if(dp->arity!=1)
	{
		semanticError(SE_TYPE_NOT_UNARY, data, vars);
		dp=&DPError;
	}

		// Reverse the order of the variables in the list vars and
		//  set the data type for every variable:
	for(dest=0, src=vars; src ;)
	{
			// Remove first variable from the 'src' list:
		cur=src;
		src=src->next;
			// Set its data type to 'dp'
		cur->type=dp;
			// Put this variable into dest list:
		cur->next=dest;
		dest=cur;
	}

		// Append 'dest' list to the list of all program variables:
		//  NOTE: 'vars' points to the last node in the list
	if(!last.var) {first.var=dest; last.var=vars;}
	else {last.var->next=dest; last.var=vars;}
}


/*
TempList *Rules::dataIdAdd(TempList *list, int symbolno)
DESCRIPTION
	Used while parsing arguments in the declaration of a program predicate
	Adds a new type of data predicate argument
NOTE
	The types of arguments in the list appear in the REVERSE ORDER

PARAMETERS
	list     -  list made of temporary nodes that contain data types
	             of program predicate arguments, can be NULL for the
	             first argument
	symbolno -  Symol number of the argument to be added to the list
RETURNS
	New list of data arguments
*/
TempList *Rules::dataIdAdd(TempList *list, int symbolno)
{
		// Allocate the new temporary list node:
	TempList *param=(TempList *)tempSrc.New();

		// Increase the arguments count
	param->size=(list)?list->size+1:1;

		// Get the pointer to the data type
	param->type=(DPnode *)((symbt.getVal(symbolno))->ptr);

		// Make sure the predicate for the data type is unary
	if(!param->type||param->type->arity!=1)
	{
			// If not, report an error:
		semanticError(SE_TYPE_NOT_UNARY, symbolno, 0);
		param->type=&DPError;  // Error recovery
	}

		// Add the node to the list and return the new list:
	param->next=list;
	return(param);
}


/*
void Rules::programPredNew(int symbolno, TempList *params, int tail)
DESCRIPTION
	The action handler that creates a new program predicate declaration
	This action creates program predicate node and adds it to the list.

PARAMETERS:
	symbolno -  Symbol number of the predicate identifier
	             (always have to have type ID)
	params   -  Temporary list of parameter types for the predicate
	             (appear in REVERSE ORDER)
	            If NULL, then the predicate has no parameters
	tail     -  If the tail for this program predicate is present, this
	             variable contains the symbol number of the data predicate
	             in the tail. Otherwise, it contains zero.
*/
void Rules::programPredNew(int symbolno, TempList *params, int tail)
{
	int i;
	unsigned int num, product;
	TempList *cur;
		// Allocate the node
	PPnode *pp=(PPnode *)permMSrc.alloc(sizeof(PPnode));

		// Set the type for this symbol number and initialize the
		//  link from the symbol table to the node.
	symbt.setType(symbolno, TYPE_PROGRAM_PRED);
	(symbt.getVal(symbolno))->ptr=pp;

		// Initialize the node:
	pp->symbolno=symbolno;
	pp->arity=0;
	pp->hornMeta=PP_NOTHING;
	pp->usage=PP_DECLARED;
	pp->params=0;
	pp->numOfAtoms=1;
	pp->tail=0;
	pp->next=0;

		// Add the node to the list of program predicates:
	if(!last.pred) last.pred=first.pred=pp;
	else last.pred=last.pred->next=pp;

		// Proceed if the predicate has any parameters:
	if(params)
	{
		pp->arity=params->size;
			// Allocate an array for the predicate parameters:
		pp->params=(DPnode**)permMSrc.alloc(pp->arity*sizeof(DPnode *));

			// Copy parameters from the list to array:
		for(i=pp->arity-1;i>=0&&params;i--)
		{
				// Save the current node and go to the next:
			cur=params;
			params=params->next;

				// Copy the type from the current
				//  node and release it:
			pp->params[i]=cur->type;
			tempSrc.Delete(cur);
		}

			// If the number of elements in the list was incorrect,
			//  there is an error in the program.
		if(i>=0||params)
		{
			fprintf(stderr, "CRITICAL ERROR: Incorrect construction"
			    " of the parameter list of a program predicate\n");
			exit(1);
		}

			// If there is a tail for this program predicate:
		if(tail>0)
		{
				// Get the pointer to the node to the tail.
			pp->tail=(DPnode *)((symbt.getVal(tail))->ptr);

				// If the arity of the tail is not equal to the
				// arity of this predicate, report an error and
				// consider this program predicate
				// to be without a tail.
			if(!pp->tail||pp->tail->arity!=pp->arity)
			{
				semanticError(SE_PP_BAD_TAIL, pp);
				pp->tail=0;
			}
		}

			// Check the number of possible combinations of
			// parameters for this predicate. If this number is
			// greater then the biggest unsigned integer, the
			// program will not be able to handle this predicate.
			// Therefore, report an error and stop.
		for(product=1,i=0;i<pp->arity;i++)
		{
				// Get the number of the values for the given
				// parameter. If that number is zero (which can
				// happen only because of error) it doesn't
				// make sense to proceed any further.
			num=(unsigned)pp->params[i]->valNum;
			if(num<=0) {pp->numOfAtoms=0; return;};

				// Check whether we would get overflow during
				// multiplication. If not, then multiply.
			if(UINTEGER_MAX/num <= product) break;
			product*=num;
		}

			// If the 'break' worked, there is an error:
		if(i<pp->arity) semanticError(SE_MANY_PPVALS, pp);
		else pp->numOfAtoms=product;
	}
}


/*
void Rules::dataNew(int symbolno, VListHandler *vlh, ExpressionHandler *eh)
DESCRIPTION
	This method can be used to create both data predicate or weight function

PARAMETERS
	symbolno  -  Symbol number of the name of the predicate or function
	vlh       -  Handler for the list of arguments
	eh        -  If this is a weight function, it contains the arithmetic
	              expression for the weight; otherwise, it should be NULL
*/
void Rules::dataNew(int symbolno, VListHandler *vlh, ExpressionHandler *eh)
{
	STVal *val=symbt.getVal(symbolno);
	WFnode *wn;
	Value *value;
	int result=0;


		// Check whether data predicate or weight function with this
		// name already exists. If not, then add it.
	if(symbt.getType(symbolno)==TYPE_NONE)
	{
		symbt.setType(symbolno, (eh)?TYPE_WEIGHT_FUN:TYPE_DATA_PRED);

			// Allocating the proper node:
		val->ptr=wn=(WFnode *)permMSrc.alloc(
			(eh)?sizeof(WFnode):sizeof(DPnode) );

			// Initialize the node:
		if(eh) wn->tmpResults=0;
		wn->arity=vlh->arity;
		wn->symbolno=symbolno;
		wn->valNum=0;
		wn->tmpArguments=0;
		wn->next=0;

			// Add the node to the data list:
		if(!last.data) first.data=last.data=wn;
		else last.data=last.data->next=wn;
	}

	wn=(WFnode *)val->ptr;

		// If this is weight function,
	if(eh)
	{
		if(eh->type==ET_NUM)
			result=eh->first->num;
		else
		{
			semanticError(SE_WEIGHT_EXP_TYPE,symbolno, eh);
			result=1;
		}
		deleteExpression(eh);

			// Indicate that the number in the variable 'result'
			//   has occurred in the program.
		strConst.registerNumber(result);
	}

		// Check whether there are any values attached to the value list
		// handler (it may happen otherwise if there was error recovery)
	if(vlh->first && vlh->last)
	{
		if(vlh->arity==wn->arity)
		{
				// Adding the arguments from the list:
			wn->valNum+=vlh->size;
			vlh->last->next=wn->tmpArguments;
			wn->tmpArguments=vlh->first;

				// Add the result if necessary:
			if(eh)
			{
					// Allocate and initialize the node:
				value=valSrc.New();
				value->type=EXPR_NUM;
				value->num=result;
				value->fileName=0; value->lineNum=0;

					// Add the value to the list:
				value->next=wn->tmpResults;
				wn->tmpResults=value;
			}
		}
		else// If the arity of the current arguments is not equal to
		    //the arity of the program predicate or weight function,
		    // report a semantic error:
			semanticError(SE_MANY_ARITIES, symbolno, vlh->arity);
	}

		// Release the node allocated for the list handler:
	tempSrc.Delete(vlh);
}


/*
VListHandler *Rules::valueNewRange(ExpressionHandler *min,
                                   ExpressionHandler *max)
DESCRIPTION
	Create a list of values for a data predicate from a range

PARAMETERS
	min   -  Handler of the expression that contains the minimum
	max   -  Handler of the expression that contains the maximum
RETURNS
	List of values
*/
VListHandler *Rules::valueNewRange(ExpressionHandler *min,
                                   ExpressionHandler *max)
{
	int nMin, nMax;
	int combinedType, i;
	Value *cur;
	char *fileName;
		// Allocate the value list handler:
	VListHandler *vlh=(VListHandler *)tempSrc.New();
		// Get the line number and the file name where the range occurrs
	int lineNum=lexan.getLineNum(isParserLookahead(), &fileName);

		// Initialize value list handler node to the empty list with
		//  arity one (by the way, the arity of a range is always one)
	vlh->arity=1;
	vlh->size=0;
	vlh->first=vlh->last=0;

		// Check whether at least one expression has the type that is
		//  not ET_NUM
	combinedType=min->type|max->type;
	if(combinedType!=ET_NUM)
	{
			// If so, report the semantic error, delete the
			//  expression and return the empty handler
		if(IS_EXP_SAY_ERROR(combinedType))
			semanticError(SE_RANGE_TYPES, min->type, max->type);
		deleteExpression(min);
		deleteExpression(max);
		return(vlh);
	}

		// Both expressions are numbers: save their values and
		//  deallocate the expression nodes:
	nMin=min->first->num;
	nMax=max->first->num;
	expSrc.Delete(min->first);
	expSrc.Delete(max->first);
	tempSrc.Delete(min);
	tempSrc.Delete(max);

		// Make sure minimum is not greater than maximum; otherwise,
		// report semantic error and return the empty list
	if(nMin>nMax)
	{
		semanticError(SE_RANGE_WRONG, nMin, nMax);
		return(vlh);
	}

		// Make sure that the range is not too big (otherwise, stupid
		//  user's error can make the program freeze by making it use
		//  too much memory and too big arrays)
	if(nMax-nMin>RANGE_UPPER_LIMIT)
	{
		semanticError(SE_RANGE_BIG, nMin, nMax);
		return(vlh);
	}

		// Indicate that the numbers occurred in the program
		//  (need to do it only for the limits)
	strConst.registerNumber(nMin);
	strConst.registerNumber(nMax);

		// Create the list itself:
	for(i=nMin; i<=nMax;i++)
	{
			// Allocate the node and add it to the list:
		if(!vlh->last) cur=vlh->first=vlh->last=valSrc.New();
		else cur=vlh->last=vlh->last->next=valSrc.New();
		vlh->size++; // Increase the node count

			// Initialize the node:
		cur->num=i;
		cur->type=EXPR_NUM;
		cur->next=0;
		cur->lineNum=lineNum;
		cur->fileName=fileName;
	}

	return(vlh); // Return the list
}


/*
VListHandler *Rules::valueAdd(VListHandler *vlh, Value *val)
DESCRIPTION
	This method is used to parse arguments of data predicates or weight
	functions in the declaration sequence. It adds a value to a list
	corresponding to a certain set of arguments.

PARAMETERS
	vlh   - Value list handler for the one set of arguments,
	         can be NULL for the first value
	val   - Value that needs to be added to the 'vlh'
RETURNS
	Handler for the updated list of values
*/
VListHandler *Rules::valueAdd(VListHandler *vlh, Value *val)
{

	if(!vlh)
	{	// If the handler is NULL, allocate and initialize it:
		vlh=(VListHandler *)tempSrc.New();
		vlh->first=vlh->last=val;
			// Value list handler that is used in this method
			//  always have the 'size' equal to one.
		vlh->size=1;
		vlh->arity=1;   // One for the first value
	}
	else
	{	// Otherwise, add the value and increase the arity
		vlh->last=vlh->last->next=val;
		vlh->arity++;
	}

	return(vlh);  // Return the handler
}


/*
Value *Rules::valueNew(int type, int v)
DESCRIPTION
	Create a value from the number or string constant

PARAMETERS
	type  -  Can be either EXPR_NUM or EXPR_STRING
	v     -  If type==EXPR_NUM, this is actual numerical value
	         If type==EXPR_STRING, this is a symbol number of
	          the string constant
RETURNS
	Newly created value node
*/
Value *Rules::valueNew(int type, int v)
{
	Value *val=valSrc.New(); // Allocate the node
	char *fileName;
		// Get the line number and file name:
	int lineNum=lexan.getLineNum(isParserLookahead(), &fileName);

		// Initialize the node:
	val->type=type;
	val->num=v;
	val->next=0;
	val->lineNum=lineNum;
	val->fileName=fileName;

	    // If this is a number, then register the number;
	    // If this is a constant, add the constant to the string constants
	if(type==EXPR_NUM) strConst.registerNumber(v);
	else
		if(type==EXPR_STRING) strConst.addConstant(v);
		else
		{
			fprintf(stderr, "CRITICAL ERROR in "
				"Rules::valueNew\n\n");
			exit(1);
		}

	return(val);  // Return the value
}


/*
ExpressionHandler *Rules::expErrorID(int symbolno)
DESCRIPTION
	This action is designed to catch the case when ID is used in expression
	and make it semantic error instead of syntax, so that the program
	can print more detailed message. The most probable cause of this is
	that some constant should have been given in the command line,
	but that was not done

PARAMETERS
	symbolno  -  Symbol number of the unknown identifier

RETURNS
	Expression handler corresponding to error expression.
*/
ExpressionHandler *Rules::expErrorID(int symbolno)
{
		// Allocate and initialize ERROR expression handler node:
	ExpressionHandler *eh=(ExpressionHandler *)tempSrc.New();
	eh->next=0;
	eh->first=eh->last=0;
	eh->type=ET_ERROR;

		// Report the semantic error and return the node:
	semanticError(SE_EXP_UNKNOWN_ID, symbt.getName(symbolno));
	return eh;
}


/*
ExpressionHandler *Rules::expNew(int number, int type)
DESCRIPTION:
	Create simple arithmetic expression for program variable, integer, or
	string constant. Creates 'Expression' and 'ExpressionHandler' nodes

PARAMETERS:
	number  - If we need to create an expression from integer, this is its
	           numerical value. For string constant or expression this
	           is a symbol number.
	type    - Type of the expression as it should appear
	           in the expression node.

RETURNS: Pointer to the 'ExpressionHandler' node for the given expression
*/
ExpressionHandler *Rules::expNew(int number, int type)
{
	ExpressionHandler *eh;
	Expression *exp;
	VarNode *var;

		// Allocate expression handler node:
	eh=(ExpressionHandler *)tempSrc.New();
	eh->next=0;

		// If a variable appears before rebuilding,
		//   instantly return an error expression.
	if(type != EXPR_NUM && !isDataFinished)
	{
		semanticError(SE_EXPR_DECLARATION, number, type);
		eh->first=eh->last=0;
		eh->type=ET_ERROR;
		return(eh);
	}

		// Initialize expression handler node:
	eh->first=eh->last=exp=expSrc.New();

		// Initialize expression node
	exp->token=type;
	exp->next=0;

		// Initialize type-specific variables:
	switch(type)
	{
	case EXPR_NUM:
		eh->type=ET_NUM;
		exp->num=number;
		break;

	case EXPR_STRING:
		eh->type=ET_STRING;
		exp->num=strConst.getNumberForConstant(number);
		break;

	case EXPR_VAR:
		var=(VarNode *)(symbt.getVal(number)->ptr);
		var->valIndex=VAR_FREE;	// Mark the variable as VAR_FREE.
		exp->var=var;

			// Check whether the variable
			//  can take string constant values.
		if(var->type->valNum >=1 &&
			!strConst.isNumber(var->type->arguments[0]))
		{
			// Some of the values are string constants. Now check
			//  whether all the values are string constants:
			if(!strConst.isNumber(
				   var->type->arguments[var->type->valNum-1]))
				eh->type=ET_STRVAR;	// If they all are
			else
				eh->type=ET_MIXVAR; // If some of them are not
		}
		else
			eh->type=ET_VAR;  // All the values are numerical
		break;

	default:
		fprintf(stderr, "CRITICAL ERROR in Rules::expNew\n"
			"wrong expression type is given as a parameter\n\n");
		exit(1);
	}

	return(eh);
}


/*
ExpressionHandler *Rules::expMerge(ExpressionHandler *HLeft,
	                           ExpressionHandler *HRight, int Operation)
DESCRIPTION:
	Given two arithmetic expressions as left and right operand and an
	operator (that can also be any function except weight function)
	creates a new arithmetic expression. It either calculates the
	resulting value, if the calculation is possible during compile-time,
	or simply merges them into one expression. If expressions cannot be
	merged, it reports semantic error and returns the error expression.

PARAMETERS:
	HLeft     -  Pointer to the expression handler for the left operand
	HRight    -  Pointer to the expression handler for the right operand
	Operation -  Code of operator. Cannot be relational operator(>,<=...)

RETURNS: Expression handler of the created expression.
*/
ExpressionHandler *Rules::expMerge(ExpressionHandler *HLeft,
                                   ExpressionHandler *HRight, int Operation)
{
	int newType;
	int rightNum=0;

		// Do this all stuff if the operation is not unary:
	if(Operation!=EXPR_ABS&&Operation!=EXPR_UMINUS)
	{
			// Calculate type of the resulting expression
		newType=HLeft->type|HRight->type;

		     // Check whether we can possibly combine these expressions:
		if(IS_EXP_NOT_COMBINE(newType))
		{
			if(IS_EXP_SAY_ERROR(newType))
				semanticError(SE_EXP_TYPES,
				              HLeft->type, HRight->type);

				// Release the whole expression
				//  from the right operand
			deleteExpression(HRight);

				// Release all the nodes from the left
				// expression and set it to be error expression
			deleteExpression(HLeft, 1);
			return(HLeft);	// Return
		}

			// If we need to combine the expressions, then append
			//  the list for the right operand to the list for the
			//  left operand. Operation node will be appended later.
		if(newType!=ET_NUM)
		{
			HLeft->last->next=HRight->first;
			HLeft->last=HRight->last;
		}
		else
		{
			rightNum=HRight->first->num;
			expSrc.Delete(HRight->first);
		}

		HLeft->type=newType; // Assigning a new type to the left operand

			// Release second operand.
		tempSrc.Delete(HRight);
	}

		// If the left expression is an error, don't do anything
	if(HLeft->type==ET_ERROR) return(HLeft);

	/*
	Situation to this point:
	    If the operation was binary:
	       1) We calculated the new type of the expression, and it was not
	          an error type.
	       2) If the expression could not be evaluated, we appended the list
	          for the right operand to the list for the left operand.

	    If the operation was unary:
	        1) We made sure that the type is not the error type.

	STILL NEED: Either evaluate the expression or append an operator node.
	*/

		// Use the type of the resulting expression to check whether
		//  we need to calculate or combine the expressions.
		//  NOTE: for unary operation, the type of the result is always
		//   equal to the type of the operand.
	if(HLeft->type==ET_NUM)
	{
		if(!expEvaluate(HLeft->first->num, rightNum, Operation,
			&(HLeft->first->num))) deleteExpression(HLeft, 1);
	}
	else
	{
			// Cannot perform the operation. Need only to append
			//  operator node to the end of the list
		HLeft->last=HLeft->last->next=expSrc.New();
		HLeft->last->next=0;
		HLeft->last->token=Operation;
	}

	return(HLeft); // Return the result.
}


/*
int Rules::checkErrorsBeforeRules()
DESCRIPTION
	This method is called after data sequence and declaration sequence
	are parsed but before rules are started. If there were no syntax or
	semantic errors before, the method returs TRUE. Otherwise, it returns
	FALSE, interrupts the lexical analyzer (so that rules won't be parsed)
	and prints the message informing the user about this

RETURNS
	TRUE or FALSE (see above)
*/
int Rules::checkErrorsBeforeRules()
{
	if(syntaxErrorCount + semanticErrorCount > 0)
	{
			// Print the message via 'semanticError' method.
		semanticError(SM_STOP_BEFORE_RULES);

			// Interrupt the input and return false.
		lexan.interruptInput();
		return 0;
	}

	return 1;
}


/*
int Rules::finishParse()
RULE: "program : dataSequence declarationSequence ruleSequence"
DESCRIPTION:
	The last action of the parser.
*/
int Rules::finishParse()
{
	reportUnusedVariables();
	reportUnusedProgramPredicates();
	if(!DEBUG_DONTFREE)
	{
		tempSrc.freeAll();
	}
		// If there were any errors or if input was interrupted for some
		//  reason, then stop grounding process at this moment.
	if(syntaxErrorCount + semanticErrorCount + lexan.getErrorCount() > 0 ||
		lexan.isInputInterrupted())
	{
			// If input has not been interrupted, print the
			//  message that grounding will be stopped.
		if(!lexan.isInputInterrupted()) semanticError(SM_INTERRUPTED);
		return 0;
	}
	else
		return 1;
}
