/*==========================================================================;
 *
 *  AUTHOR:     Subash, Agila
 *
 *  FILE NAME: symtable.cpp
 *
 *  CONTENTS: Implementation of the class 'SymbolTable'.
 *
 *  DATE:     March 19, 2004
 ***************************************************************************/

#include <string.h>
#include "symtable.h"

	// Basic structure of the node table:
struct STNodeType
{
	char *name;   // Pointer to the string entered into the symbol table
	int next;     // Reference to the next node in the list: used to resolve
	              // hash table clashes
	int type;     // Type of the entry (default TYPE_NONE)
	union STVal val; // Value of the entry

};

//-----------------constructor---------------------------------
SymbolTable::SymbolTable():
	mem(INITIAL_SPELL_SIZE,SPELL_GROWTH_FACTOR,MAX_SPELL_SIZE)
{
	int i;
	cntNTEntry= 0;
	hashTableSize = MAX_HT_ELEMENTS;
	nodeTableSize = MAX_NT_ELEMENTS;
	hashTable = (int *)malloc(hashTableSize * sizeof (int));
	STNode = (struct STNodeType *)
		malloc(nodeTableSize * sizeof(struct STNodeType));

	if (STNode == NULL || hashTable == NULL)
	{
		fprintf(stderr, "CRITICAL ERROR:Symbol Table : "
			"Error Allocating Node Table or Hash Table\n");
		exit(1);
	}

	for (i=0;i<MAX_HT_ELEMENTS;i++)
	{
		hashTable[i] =  ST_FREE;
	}

}
//-----------------Destructor-----------------------------
SymbolTable::~SymbolTable()
{
	mem.freeAll();
	free(hashTable);
	free(STNode);
}

//-------------Reallocate the node table -------------------

void SymbolTable::NTRealloc()
{
	nodeTableSize = (int)((double)nodeTableSize * NODE_GROWTH_FACTOR);
	STNode = (struct STNodeType *)
		realloc(STNode,nodeTableSize * sizeof(struct STNodeType));
	if (STNode == NULL)
	{
		fprintf(stderr, "CRITICAL ERROR: Symbol Table :"
			" Error Reallocating Node Table\n");
		exit(1);
	}
}
//--------------Reallocate the hash table --------------------
void SymbolTable::HTRealloc()
{
	int i;
	hashTableSize = (int)((double) hashTableSize * HASH_GROWTH_FACTOR);
	hashTable = (int *)realloc(hashTable, hashTableSize * sizeof (int));


	if (hashTable == NULL)
	{
		printf("CRITICAL ERROR: Symbol Table : "
			"Error allocating Hash Table\n");
		exit(1);
	}

	for (i=0;i<hashTableSize;i++)
		hashTable[i] =  ST_FREE;


	//set all the next fields of the node table to ST_FREE
	for(i=1;i<=cntNTEntry;i++)
		STNode[i].next = ST_FREE;


	// Re-insert all the nodes into the hash table:
	for(i=1;i<=cntNTEntry;i++)
	{
		//find the hash value
		int indexVal = calHashValue(STNode[i].name);

	//if hash table value is ST_FREE,assign hashTable[hash value] = i;
		if(hashTable[indexVal] == ST_FREE)
			hashTable[indexVal] =i;
		else
		{
			int currIndex = hashTable[indexVal];

			while (STNode[currIndex].next != ST_FREE)
				currIndex = STNode[currIndex].next;
			STNode[currIndex].next = i;
		}

	}
}

//--------------------Insert Name --------------------

int SymbolTable::insertName(char* name)
{

	if(name == NULL)
	{
		fprintf(stderr, "CRITICAL ERROR: Symbol Table :"
			" input is a null string\n");
		exit(1);
	}

	//get the hash value for the string parameter
	int indexVal ;
	int currIndex , oldIndex;


		//Reallocate Hash Table if needed
	if (HASH_REALLOCATE_RATIO <= ((float)cntNTEntry/(float)hashTableSize))
		HTRealloc();

	indexVal= calHashValue(name);


	// Insert the name into the hash table:
	// if the hash table value is not ST_FREE,loop through
	if (hashTable[indexVal] != ST_FREE)
	{
		 oldIndex=currIndex = hashTable[indexVal];

		 while (currIndex != ST_FREE)
		 {
			if(strcmp (STNode[currIndex].name,name)==0)
				return currIndex;
			oldIndex=currIndex;
			currIndex = STNode[currIndex].next;
		}

		cntNTEntry++;
		STNode[oldIndex].next = cntNTEntry;
	}
	else
	{
		cntNTEntry++;
		hashTable[indexVal] = cntNTEntry;
	}


		//Reallocate Node Table if needed
	if(cntNTEntry>=nodeTableSize)
		NTRealloc();

	     // Allocate the memory for the identifier and copy it into there:
	STNode[cntNTEntry].name = (char*) mem.alloc(strlen(name) + 1);
	strcpy(STNode[cntNTEntry].name ,name);

		// Initialize the remaining fields of the node:
	STNode[cntNTEntry].next = ST_FREE;
	STNode[cntNTEntry].type = TYPE_NONE;
	STNode[cntNTEntry].val.ptr=0;

	return cntNTEntry;

}

//------------------ Calculate hash value -------------------------

int SymbolTable::calHashValue(char* name1)
{
	int hashval;
	for(hashval = 0;*name1 != '\0';name1++)
		hashval += *name1;
	return hashval % hashTableSize;


}

//------------------------getName--------------------------------------


char* SymbolTable::getName(int num)
{
	if((num <= 0) || (num > cntNTEntry))
	{
		fprintf(stderr, "CRITICAL ERROR: Symbol Table :"
			" Accesing out of memory\n");
		return ("");
	}



	return (STNode[num].name );


}

//-------------------getType----------------------------

int SymbolTable::getType(int num)
{

	if((num <= 0) || (num > cntNTEntry))
	{
		fprintf(stderr, "CRITICAL ERROR: Symbol Table : "
			"Accesing out of memory\n");
		return 0;
	}

	return (STNode[num].type );

}

//-------------------setType-----------------------------
void SymbolTable::setType(int num,int type1)
{

	if((num <= 0) || (num > cntNTEntry))
	{
		fprintf(stderr, "CRITICAL ERROR: Symbol Table :"
			" Accesing out of memory\n");
		return ;
	}

	STNode[num].type = type1;


}

//--------------getVal-------------------------------

STVal* SymbolTable::getVal(int num)
{

	if((num <= 0) || (num > cntNTEntry))
	{
		fprintf(stderr, "CRITICAL ERROR: Symbol Table :"
			" Accesing out of memory\n");
		return 0;
	}
	union STVal *temp1;
	temp1 = & STNode[num].val;
	return (temp1);
}

  //------------clear--------------------------
void SymbolTable::clear()
{
		// Release all memory:
	mem.freeAll();
	free(hashTable);
	free(STNode);

		// Do the same thing as the constructor:
	cntNTEntry= 0;
	hashTableSize = MAX_HT_ELEMENTS;
	nodeTableSize = MAX_NT_ELEMENTS;
	hashTable = (int *)malloc(hashTableSize * sizeof (int));
	STNode = (struct STNodeType *)
		malloc(nodeTableSize * sizeof(struct STNodeType));

	if (STNode == NULL || hashTable == NULL)
	{
		fprintf(stderr, "CRITICAL ERROR:Symbol Table : "
			"Error Allocating Node Table or Hash Table\n");
		exit(1);
	}

	for (int i=0;i<MAX_HT_ELEMENTS;i++)
	{
		hashTable[i] =  ST_FREE;
	}
}
//------------------print-------------------------------
void SymbolTable::print()
{
	int i;
		//for debug purpose
	for(i=1;i<=cntNTEntry;i++)
		printf("%s\n", STNode[i].name);
}

// Return the range of the symbol numbers used.
// The first returned symbol number is always 1, the last is equal
//  to the 'cntNTEntry'
void SymbolTable::getNumRange(int *min, int *max)
{
	if(min) *min=1;
	if(max) *max=cntNTEntry;
}
